#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2020, Google Inc.
#
# Author: Paul Elder <paul.elder@ideasonboard.com>
#
# v4l2_compat_test.py - Test the V4L2 compatibility layer

import argparse
import glob
import os
from packaging import version
import re
import shutil
import signal
import subprocess
import sys

MIN_V4L_UTILS_VERSION = version.parse("1.21.0")

TestPass = 0
TestFail = -1
TestSkip = 77


supported_pipelines = [
    'bcm2835-isp',
    'uvcvideo',
    'vimc',
]


def grep(exp, arr):
    return [s for s in arr if re.search(exp, s)]


def run_with_stdout(*args, env={}):
    try:
        with open(os.devnull, 'w') as devnull:
            output = subprocess.check_output(args, env=env, stderr=devnull)
        ret = 0
    except subprocess.CalledProcessError as err:
        output = err.output
        ret = err.returncode
    return ret, output.decode('utf-8').split('\n')


def extract_result(result):
    res = result.split(', ')
    ret = {}
    ret['total']     = int(res[0].split(': ')[-1])
    ret['succeeded'] = int(res[1].split(': ')[-1])
    ret['failed']    = int(res[2].split(': ')[-1])
    ret['warnings']  = int(res[3].split(': ')[-1])
    ret['device']    = res[0].split()[4].strip(':')
    ret['driver']    = res[0].split()[2]
    return ret


def test_v4l2_compliance(v4l2_compliance, ld_preload, device, base_driver):
    ret, output = run_with_stdout(v4l2_compliance, '-s', '-d', device, env={'LD_PRELOAD': ld_preload})
    if ret < 0:
        output.append(f'Test for {device} terminated due to signal {signal.Signals(-ret).name}')
        return TestFail, output

    result = extract_result(output[-2])
    if result['failed'] == 0:
        return TestPass, output

    # vimc will fail s_fmt because it only supports framesizes that are
    # multiples of 3
    if base_driver == 'vimc' and result['failed'] == 1:
        failures = grep('fail', output)
        if re.search('S_FMT cannot handle an invalid format', failures[0]) is None:
            return TestFail, output
        return TestPass, output

    return TestFail, output


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--all', action='store_true',
                        help='Test all available cameras')
    parser.add_argument('-s', '--sanitizer', type=str,
                        help='Path to the address sanitizer (ASan) runtime')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Make the output verbose')
    parser.add_argument('v4l2_compat', type=str,
                        help='Path to v4l2-compat.so')
    args = parser.parse_args(argv[1:])

    # Compute the LD_PRELOAD value by first loading ASan (if specified) and
    # then the V4L2 compat layer.
    ld_preload = []
    if args.sanitizer:
        ld_preload.append(args.sanitizer)
    ld_preload.append(args.v4l2_compat)
    ld_preload = ':'.join(ld_preload)

    v4l2_compliance = shutil.which('v4l2-compliance')
    if v4l2_compliance is None:
        print('v4l2-compliance is not available')
        return TestSkip

    ret, out = run_with_stdout(v4l2_compliance, '--version')
    if ret != 0 or version.parse(out[0].split()[1].replace(',', '')) < MIN_V4L_UTILS_VERSION:
        print('v4l2-compliance version >= 1.21.0 required')
        return TestSkip

    v4l2_ctl = shutil.which('v4l2-ctl')
    if v4l2_ctl is None:
        print('v4l2-ctl is not available')
        return TestSkip

    ret, out = run_with_stdout(v4l2_ctl, '--version')
    if ret != 0 or version.parse(out[0].split()[-1]) < MIN_V4L_UTILS_VERSION:
        print('v4l2-ctl version >= 1.21.0 required')
        return TestSkip

    dev_nodes = glob.glob('/dev/video*')
    if len(dev_nodes) == 0:
        print('no video nodes available to test with')
        return TestSkip

    failed = []
    drivers_tested = {}
    for device in dev_nodes:
        ret, out = run_with_stdout(v4l2_ctl, '-D', '-d', device, env={'LD_PRELOAD': ld_preload})
        if ret < 0:
            failed.append(device)
            print(f'v4l2-ctl failed on {device} with v4l2-compat')
            continue
        driver = grep('Driver name', out)[0].split(':')[-1].strip()
        if driver != "libcamera":
            continue

        ret, out = run_with_stdout(v4l2_ctl, '-D', '-d', device)
        if ret < 0:
            failed.append(device)
            print(f'v4l2-ctl failed on {device} without v4l2-compat')
            continue
        driver = grep('Driver name', out)[0].split(':')[-1].strip()
        if driver not in supported_pipelines:
            continue

        # TODO: Add kernel version check when vimc supports scaling
        if driver == "vimc":
            continue

        if not args.all and driver in drivers_tested:
            continue

        print(f'Testing {device} with {driver} driver... ', end='')
        ret, msg = test_v4l2_compliance(v4l2_compliance, ld_preload, device, driver)
        if ret == TestFail:
            failed.append(device)
            print('failed')
        else:
            print('success')

        if ret == TestFail or args.verbose:
            print('\n'.join(msg))

        drivers_tested[driver] = True

    if len(drivers_tested) == 0:
        print(f'No compatible drivers found')
        return TestSkip

    if len(failed) > 0:
        print(f'Failed {len(failed)} tests:')
        for device in failed:
            print(f'- {device}')

    return TestPass if not failed else TestFail


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# Copyright 2022 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Validate RequireContext and AllowedContext annotations before generation."""

import mojom.generate.check as check
import mojom.generate.module as module


class Check(check.Check):
  def __init__(self, *args, **kwargs):
    self.kind_to_interfaces = dict()
    super(Check, self).__init__(*args, **kwargs)

  def _IsPassedInterface(self, candidate):
    if isinstance(
        candidate.kind,
        (module.PendingReceiver, module.PendingRemote,
         module.PendingAssociatedReceiver, module.PendingAssociatedRemote)):
      return True
    return False

  def _CheckInterface(self, method, param):
    # |param| is a pending_x<Interface> so need .kind.kind to get Interface.
    interface = param.kind.kind
    if interface.require_context:
      if method.allowed_context is None:
        raise check.CheckException(
            self.module, "method `{}` has parameter `{}` which passes interface"
            " `{}` that requires an AllowedContext annotation but none exists.".
            format(
                method.mojom_name,
                param.mojom_name,
                interface.mojom_name,
            ))
      # If a string was provided, or if an enum was not imported, this will
      # be a string and we cannot validate that it is in range.
      if not isinstance(method.allowed_context, module.EnumValue):
        raise check.CheckException(
            self.module,
            "method `{}` has AllowedContext={} which is not a valid enum value."
            .format(method.mojom_name, method.allowed_context))
      # EnumValue must be from the same enum to be compared.
      if interface.require_context.enum != method.allowed_context.enum:
        raise check.CheckException(
            self.module, "method `{}` has parameter `{}` which passes interface"
            " `{}` that requires AllowedContext={} but one of kind `{}` was "
            "provided.".format(
                method.mojom_name,
                param.mojom_name,
                interface.mojom_name,
                interface.require_context.enum,
                method.allowed_context.enum,
            ))
      # RestrictContext enums have most privileged field first (lowest value).
      interface_value = interface.require_context.field.numeric_value
      method_value = method.allowed_context.field.numeric_value
      if interface_value < method_value:
        raise check.CheckException(
            self.module, "RequireContext={} > AllowedContext={} for method "
            "`{}` which passes interface `{}`.".format(
                interface.require_context.GetSpec(),
                method.allowed_context.GetSpec(), method.mojom_name,
                interface.mojom_name))
      return True

  def _GatherReferencedInterfaces(self, field):
    key = field.kind.spec
    # structs/unions can nest themselves so we need to bookkeep.
    if not key in self.kind_to_interfaces:
      # Might reference ourselves so have to create the list first.
      self.kind_to_interfaces[key] = set()
      for param in field.kind.fields:
        if self._IsPassedInterface(param):
          self.kind_to_interfaces[key].add(param)
        elif isinstance(param.kind, (module.Struct, module.Union)):
          for iface in self._GatherReferencedInterfaces(param):
            self.kind_to_interfaces[key].add(iface)
    return self.kind_to_interfaces[key]

  def _CheckParams(self, method, params):
    # Note: we have to repeat _CheckParams for each method as each might have
    # different AllowedContext= attributes. We cannot memoize this function,
    # but can do so for gathering referenced interfaces as their RequireContext
    # attributes do not change.
    for param in params:
      if self._IsPassedInterface(param):
        self._CheckInterface(method, param)
      elif isinstance(param.kind, (module.Struct, module.Union)):
        for interface in self._GatherReferencedInterfaces(param):
          self._CheckInterface(method, interface)

  def _CheckMethod(self, method):
    if method.parameters:
      self._CheckParams(method, method.parameters)
    if method.response_parameters:
      self._CheckParams(method, method.response_parameters)

  def CheckModule(self):
    for interface in self.module.interfaces:
      for method in interface.methods:
        self._CheckMethod(method)



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# generator.py - Base class for a generator to convert dict to tuning file

from pathlib import Path


class Generator(object):
    def __init__(self):
        pass

    def write(self, output_path: Path, output_dict: dict, output_order: list):
        raise NotImplementedError



#!/usr/bin/env python3
#
# SPDX-License-Identifier: BSD-2-Clause
#
# Script to convert version 1.0 Raspberry Pi camera tuning files to version 2.0.
#
# Copyright 2022 Raspberry Pi Ltd

import argparse
import json
import sys

from ctt_pretty_print_json import pretty_print


def convert_v2(in_json: dict) -> str:

    if 'version' in in_json.keys() and in_json['version'] != 1.0:
        print(f'The JSON config reports version {in_json["version"]} that is incompatible with this tool.')
        sys.exit(-1)

    converted = {
        'version': 2.0,
        'target': 'bcm2835',
        'algorithms': [{algo: config} for algo, config in in_json.items()]
    }

    return pretty_print(converted)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, description=
                    'Convert the format of the Raspberry Pi camera tuning file from v1.0 to v2.0.\n')
    parser.add_argument('input', type=str, help='Input tuning file.')
    parser.add_argument('output', type=str, nargs='?',
                        help='Output converted tuning file. If not provided, the input file will be updated in-place.',
                        default=None)
    args = parser.parse_args()

    with open(args.input, 'r') as f:
        in_json = json.load(f)

    out_json = convert_v2(in_json)

    with open(args.output if args.output is not None else args.input, 'w') as f:
        f.write(out_json)



# Copyright 2022 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import unittest

import mojom.generate.check as check
from mojom_bindings_generator import LoadChecks, _Generate
from mojom_parser_test_case import MojomParserTestCase


class FakeArgs:
  """Fakes args to _Generate - intention is to do just enough to run checks"""

  def __init__(self, tester, files=None):
    """ `tester` is MojomParserTestCase for paths.
        `files` will have tester path added."""
    self.checks_string = 'attributes'
    self.depth = tester.GetPath('')
    self.filelist = None
    self.filename = [tester.GetPath(x) for x in files]
    self.gen_directories = tester.GetPath('gen')
    self.generators_string = ''
    self.import_directories = []
    self.output_dir = tester.GetPath('out')
    self.scrambled_message_id_salt_paths = None
    self.typemaps = []
    self.variant = 'none'


class MojoBindingsCheckTest(MojomParserTestCase):
  def _ParseAndGenerate(self, mojoms):
    self.ParseMojoms(mojoms)
    args = FakeArgs(self, files=mojoms)
    _Generate(args, {})

  def _testValid(self, filename, content):
    self.WriteFile(filename, content)
    self._ParseAndGenerate([filename])

  def _testThrows(self, filename, content, regexp):
    mojoms = []
    self.WriteFile(filename, content)
    mojoms.append(filename)
    with self.assertRaisesRegexp(check.CheckException, regexp):
      self._ParseAndGenerate(mojoms)

  def testLoads(self):
    """Validate that the check is registered under the expected name."""
    check_modules = LoadChecks('attributes')
    self.assertTrue(check_modules['attributes'])

  def testNoAnnotations(self):
    # Undecorated mojom should be fine.
    self._testValid(
        "a.mojom", """
      module a;
      struct Bar { int32 a; };
      enum Hello { kValue };
      union Thingy { Bar b; Hello hi; };
      interface Foo {
        Foo(int32 a, Hello hi, Thingy t) => (Bar b);
      };
    """)

  def testValidAnnotations(self):
    # Obviously this is meaningless and won't generate, but it should pass
    # the attribute check's validation.
    self._testValid(
        "a.mojom", """
      [JavaConstantsClassName="FakeClass",JavaPackage="org.chromium.Fake"]
      module a;
      [Stable, Extensible]
      enum Hello { [Default] kValue, kValue2, [MinVersion=2] kValue3 };
      [Native]
      enum NativeEnum {};
      [Stable,Extensible]
      union Thingy { Bar b; [Default]int32 c; Hello hi; };

      [Stable,RenamedFrom="module.other.Foo",
       Uuid="4C178401-4B07-4C2E-9255-5401A943D0C7"]
      struct Structure { Hello hi; };

      [ServiceSandbox=Hello.kValue,RequireContext=Hello.kValue,Stable,
       Uuid="2F17D7DD-865A-4B1C-9394-9C94E035E82F"]
      interface Foo {
        [AllowedContext=Hello.kValue]
        Foo@0(int32 a) => (int32 b);
        [MinVersion=2,Sync,UnlimitedSize,NoInterrupt]
        Bar@1(int32 b, [MinVersion=2]Structure? s) => (bool c);
      };

      [RuntimeFeature=test.mojom.FeatureName]
      interface FooFeatureControlled {};

      interface FooMethodFeatureControlled {
        [RuntimeFeature=test.mojom.FeatureName]
        MethodWithFeature() => (bool c);
      };
    """)

  def testWrongModuleStable(self):
    contents = """
      // err: module cannot be Stable
      [Stable]
      module a;
      enum Hello { kValue, kValue2, kValue3 };
      enum NativeEnum {};
      struct Structure { Hello hi; };

      interface Foo {
        Foo(int32 a) => (int32 b);
        Bar(int32 b, Structure? s) => (bool c);
      };
    """
    self._testThrows('b.mojom', contents,
                     'attribute Stable not allowed on module')

  def testWrongEnumDefault(self):
    contents = """
      module a;
      // err: default should go on EnumValue not Enum.
      [Default=kValue]
      enum Hello { kValue, kValue2, kValue3 };
      enum NativeEnum {};
      struct Structure { Hello hi; };

      interface Foo {
        Foo(int32 a) => (int32 b);
        Bar(int32 b, Structure? s) => (bool c);
      };
    """
    self._testThrows('b.mojom', contents,
                     'attribute Default not allowed on enum')

  def testWrongStructMinVersion(self):
    contents = """
      module a;
      enum Hello { kValue, kValue2, kValue3 };
      enum NativeEnum {};
      // err: struct cannot have MinVersion.
      [MinVersion=2]
      struct Structure { Hello hi; };

      interface Foo {
        Foo(int32 a) => (int32 b);
        Bar(int32 b, Structure? s) => (bool c);
      };
    """
    self._testThrows('b.mojom', contents,
                     'attribute MinVersion not allowed on struct')

  def testWrongMethodRequireContext(self):
    contents = """
      module a;
      enum Hello { kValue, kValue2, kValue3 };
      enum NativeEnum {};
      struct Structure { Hello hi; };

      interface Foo {
        // err: RequireContext is for interfaces.
        [RequireContext=Hello.kValue]
        Foo(int32 a) => (int32 b);
        Bar(int32 b, Structure? s) => (bool c);
      };
    """
    self._testThrows('b.mojom', contents,
                     'RequireContext not allowed on method')

  def testWrongMethodRequireContext(self):
    # crbug.com/1230122
    contents = """
      module a;
      interface Foo {
        // err: sync not Sync.
        [sync]
        Foo(int32 a) => (int32 b);
      };
    """
    self._testThrows('b.mojom', contents,
                     'attribute sync not allowed.*Did you mean: Sync')

  def testStableExtensibleEnum(self):
    # crbug.com/1193875
    contents = """
      module a;
      [Stable]
      enum Foo {
        kDefaultVal,
        kOtherVal = 2,
      };
    """
    self._testThrows('a.mojom', contents,
                     'Extensible.*?required.*?Stable.*?enum')



# Copyright 2014 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Generates a syntax tree from a Mojo IDL file."""

# Breaking parser stanzas is unhelpful so allow longer lines.
# pylint: disable=line-too-long

import os.path
import sys

from mojom import fileutil
from mojom.error import Error
from mojom.parse import ast
from mojom.parse.lexer import Lexer

fileutil.AddLocalRepoThirdPartyDirToModulePath()
from ply import lex
from ply import yacc

_MAX_ORDINAL_VALUE = 0xffffffff
_MAX_ARRAY_SIZE = 0xffffffff


class ParseError(Error):
  """Class for errors from the parser."""

  def __init__(self, filename, message, lineno=None, snippet=None):
    Error.__init__(
        self,
        filename,
        message,
        lineno=lineno,
        addenda=([snippet] if snippet else None))


# We have methods which look like they could be functions:
# pylint: disable=R0201
class Parser:
  def __init__(self, lexer, source, filename):
    self.tokens = lexer.tokens
    self.source = source
    self.filename = filename

  # Names of functions
  #
  # In general, we name functions after the left-hand-side of the rule(s) that
  # they handle. E.g., |p_foo_bar| for a rule |foo_bar : ...|.
  #
  # There may be multiple functions handling rules for the same left-hand-side;
  # then we name the functions |p_foo_bar_N| (for left-hand-side |foo_bar|),
  # where N is a number (numbered starting from 1). Note that using multiple
  # functions is actually more efficient than having single functions handle
  # multiple rules (and, e.g., distinguishing them by examining |len(p)|).
  #
  # It's also possible to have a function handling multiple rules with different
  # left-hand-sides. We do not do this.
  #
  # See http://www.dabeaz.com/ply/ply.html#ply_nn25 for more details.

  # TODO(vtl): Get rid of the braces in the module "statement". (Consider
  # renaming "module" -> "package".) Then we'll be able to have a single rule
  # for root (by making module "optional").
  def p_root_1(self, p):
    """root : """
    p[0] = ast.Mojom(None, ast.ImportList(), [])

  def p_root_2(self, p):
    """root : root module"""
    if p[1].module is not None:
      raise ParseError(
          self.filename,
          "Multiple \"module\" statements not allowed:",
          p[2].lineno,
          snippet=self._GetSnippet(p[2].lineno))
    if p[1].import_list.items or p[1].definition_list:
      raise ParseError(
          self.filename,
          "\"module\" statements must precede imports and definitions:",
          p[2].lineno,
          snippet=self._GetSnippet(p[2].lineno))
    p[0] = p[1]
    p[0].module = p[2]

  def p_root_3(self, p):
    """root : root import"""
    if p[1].definition_list:
      raise ParseError(
          self.filename,
          "\"import\" statements must precede definitions:",
          p[2].lineno,
          snippet=self._GetSnippet(p[2].lineno))
    p[0] = p[1]
    p[0].import_list.Append(p[2])

  def p_root_4(self, p):
    """root : root definition"""
    p[0] = p[1]
    p[0].definition_list.append(p[2])

  def p_import(self, p):
    """import : attribute_section IMPORT STRING_LITERAL SEMI"""
    # 'eval' the literal to strip the quotes.
    # TODO(vtl): This eval is dubious. We should unquote/unescape ourselves.
    p[0] = ast.Import(
        p[1], eval(p[3]), filename=self.filename, lineno=p.lineno(2))

  def p_module(self, p):
    """module : attribute_section MODULE identifier_wrapped SEMI"""
    p[0] = ast.Module(p[3], p[1], filename=self.filename, lineno=p.lineno(2))

  def p_definition(self, p):
    """definition : struct
                  | union
                  | interface
                  | enum
                  | const
                  | feature"""
    p[0] = p[1]

  def p_attribute_section_1(self, p):
    """attribute_section : """
    p[0] = None

  def p_attribute_section_2(self, p):
    """attribute_section : LBRACKET attribute_list RBRACKET"""
    p[0] = p[2]

  def p_attribute_list_1(self, p):
    """attribute_list : """
    p[0] = ast.AttributeList()

  def p_attribute_list_2(self, p):
    """attribute_list : nonempty_attribute_list"""
    p[0] = p[1]

  def p_nonempty_attribute_list_1(self, p):
    """nonempty_attribute_list : attribute"""
    p[0] = ast.AttributeList(p[1])

  def p_nonempty_attribute_list_2(self, p):
    """nonempty_attribute_list : nonempty_attribute_list COMMA attribute"""
    p[0] = p[1]
    p[0].Append(p[3])

  def p_attribute_1(self, p):
    """attribute : name_wrapped EQUALS identifier_wrapped"""
    p[0] = ast.Attribute(p[1],
                         p[3][1],
                         filename=self.filename,
                         lineno=p.lineno(1))

  def p_attribute_2(self, p):
    """attribute : name_wrapped EQUALS evaled_literal
                 | name_wrapped EQUALS name_wrapped"""
    p[0] = ast.Attribute(p[1], p[3], filename=self.filename, lineno=p.lineno(1))

  def p_attribute_3(self, p):
    """attribute : name_wrapped"""
    p[0] = ast.Attribute(p[1], True, filename=self.filename, lineno=p.lineno(1))

  def p_evaled_literal(self, p):
    """evaled_literal : literal"""
    # 'eval' the literal to strip the quotes. Handle keywords "true" and "false"
    # specially since they cannot directly be evaluated to python boolean
    # values.
    if p[1] == "true":
      p[0] = True
    elif p[1] == "false":
      p[0] = False
    else:
      p[0] = eval(p[1])

  def p_struct_1(self, p):
    """struct : attribute_section STRUCT name_wrapped LBRACE struct_body RBRACE SEMI"""
    p[0] = ast.Struct(p[3], p[1], p[5])

  def p_struct_2(self, p):
    """struct : attribute_section STRUCT name_wrapped SEMI"""
    p[0] = ast.Struct(p[3], p[1], None)

  def p_struct_body_1(self, p):
    """struct_body : """
    p[0] = ast.StructBody()

  def p_struct_body_2(self, p):
    """struct_body : struct_body const
                   | struct_body enum
                   | struct_body struct_field"""
    p[0] = p[1]
    p[0].Append(p[2])

  def p_struct_field(self, p):
    """struct_field : attribute_section typename name_wrapped ordinal default SEMI"""
    p[0] = ast.StructField(p[3], p[1], p[4], p[2], p[5])

  def p_feature(self, p):
    """feature : attribute_section FEATURE NAME LBRACE feature_body RBRACE SEMI"""
    p[0] = ast.Feature(p[3], p[1], p[5])

  def p_feature_body_1(self, p):
    """feature_body : """
    p[0] = ast.FeatureBody()

  def p_feature_body_2(self, p):
    """feature_body : feature_body const"""
    p[0] = p[1]
    p[0].Append(p[2])

  def p_union(self, p):
    """union : attribute_section UNION name_wrapped LBRACE union_body RBRACE SEMI"""
    p[0] = ast.Union(p[3], p[1], p[5])

  def p_union_body_1(self, p):
    """union_body : """
    p[0] = ast.UnionBody()

  def p_union_body_2(self, p):
    """union_body : union_body union_field"""
    p[0] = p[1]
    p[1].Append(p[2])

  def p_union_field(self, p):
    """union_field : attribute_section typename name_wrapped ordinal SEMI"""
    p[0] = ast.UnionField(p[3], p[1], p[4], p[2])

  def p_default_1(self, p):
    """default : """
    p[0] = None

  def p_default_2(self, p):
    """default : EQUALS constant"""
    p[0] = p[2]

  def p_interface(self, p):
    """interface : attribute_section INTERFACE name_wrapped LBRACE interface_body RBRACE SEMI"""
    p[0] = ast.Interface(p[3], p[1], p[5])

  def p_interface_body_1(self, p):
    """interface_body : """
    p[0] = ast.InterfaceBody()

  def p_interface_body_2(self, p):
    """interface_body : interface_body const
                      | interface_body enum
                      | interface_body method"""
    p[0] = p[1]
    p[0].Append(p[2])

  def p_response_1(self, p):
    """response : """
    p[0] = None

  def p_response_2(self, p):
    """response : RESPONSE LPAREN parameter_list RPAREN"""
    p[0] = p[3]

  def p_method(self, p):
    """method : attribute_section name_wrapped ordinal LPAREN parameter_list RPAREN response SEMI"""
    p[0] = ast.Method(p[2], p[1], p[3], p[5], p[7])

  def p_parameter_list_1(self, p):
    """parameter_list : """
    p[0] = ast.ParameterList()

  def p_parameter_list_2(self, p):
    """parameter_list : nonempty_parameter_list"""
    p[0] = p[1]

  def p_nonempty_parameter_list_1(self, p):
    """nonempty_parameter_list : parameter"""
    p[0] = ast.ParameterList(p[1])

  def p_nonempty_parameter_list_2(self, p):
    """nonempty_parameter_list : nonempty_parameter_list COMMA parameter"""
    p[0] = p[1]
    p[0].Append(p[3])

  def p_parameter(self, p):
    """parameter : attribute_section typename name_wrapped ordinal"""
    p[0] = ast.Parameter(
        p[3], p[1], p[4], p[2], filename=self.filename, lineno=p.lineno(3))

  def p_typename(self, p):
    """typename : nonnullable_typename QSTN
                | nonnullable_typename"""
    if len(p) == 2:
      p[0] = p[1]
    else:
      p[0] = p[1] + "?"

  def p_nonnullable_typename(self, p):
    """nonnullable_typename : basictypename
                            | array
                            | fixed_array
                            | associative_array"""
    p[0] = p[1]

  def p_basictypename(self, p):
    """basictypename : remotetype
                     | receivertype
                     | associatedremotetype
                     | associatedreceivertype
                     | identifier
                     | ASSOCIATED identifier
                     | handletype"""
    if len(p) == 2:
      p[0] = p[1]
    else:
      p[0] = "asso<" + p[2] + ">"

  def p_remotetype(self, p):
    """remotetype : PENDING_REMOTE LANGLE identifier RANGLE"""
    p[0] = "rmt<%s>" % p[3]

  def p_receivertype(self, p):
    """receivertype : PENDING_RECEIVER LANGLE identifier RANGLE"""
    p[0] = "rcv<%s>" % p[3]

  def p_associatedremotetype(self, p):
    """associatedremotetype : PENDING_ASSOCIATED_REMOTE LANGLE identifier RANGLE"""
    p[0] = "rma<%s>" % p[3]

  def p_associatedreceivertype(self, p):
    """associatedreceivertype : PENDING_ASSOCIATED_RECEIVER LANGLE identifier RANGLE"""
    p[0] = "rca<%s>" % p[3]

  def p_handletype(self, p):
    """handletype : HANDLE
                  | HANDLE LANGLE name_wrapped RANGLE"""
    if len(p) == 2:
      p[0] = p[1]
    else:
      if p[3] not in ('data_pipe_consumer', 'data_pipe_producer',
                      'message_pipe', 'shared_buffer', 'platform'):
        # Note: We don't enable tracking of line numbers for everything, so we
        # can't use |p.lineno(3)|.
        raise ParseError(
            self.filename,
            "Invalid handle type %r:" % p[3],
            lineno=p.lineno(1),
            snippet=self._GetSnippet(p.lineno(1)))
      p[0] = "handle<" + p[3] + ">"

  def p_array(self, p):
    """array : ARRAY LANGLE typename RANGLE"""
    p[0] = p[3] + "[]"

  def p_fixed_array(self, p):
    """fixed_array : ARRAY LANGLE typename COMMA INT_CONST_DEC RANGLE"""
    value = int(p[5])
    if value == 0 or value > _MAX_ARRAY_SIZE:
      raise ParseError(
          self.filename,
          "Fixed array size %d invalid:" % value,
          lineno=p.lineno(5),
          snippet=self._GetSnippet(p.lineno(5)))
    p[0] = p[3] + "[" + p[5] + "]"

  def p_associative_array(self, p):
    """associative_array : MAP LANGLE identifier COMMA typename RANGLE"""
    p[0] = p[5] + "{" + p[3] + "}"

  def p_ordinal_1(self, p):
    """ordinal : """
    p[0] = None

  def p_ordinal_2(self, p):
    """ordinal : ORDINAL"""
    value = int(p[1][1:])
    if value > _MAX_ORDINAL_VALUE:
      raise ParseError(
          self.filename,
          "Ordinal value %d too large:" % value,
          lineno=p.lineno(1),
          snippet=self._GetSnippet(p.lineno(1)))
    p[0] = ast.Ordinal(value, filename=self.filename, lineno=p.lineno(1))

  def p_enum_1(self, p):
    """enum : attribute_section ENUM name_wrapped LBRACE enum_value_list RBRACE SEMI
            | attribute_section ENUM name_wrapped LBRACE \
                    nonempty_enum_value_list COMMA RBRACE SEMI"""
    p[0] = ast.Enum(
        p[3], p[1], p[5], filename=self.filename, lineno=p.lineno(2))

  def p_enum_2(self, p):
    """enum : attribute_section ENUM name_wrapped SEMI"""
    p[0] = ast.Enum(
        p[3], p[1], None, filename=self.filename, lineno=p.lineno(2))

  def p_enum_value_list_1(self, p):
    """enum_value_list : """
    p[0] = ast.EnumValueList()

  def p_enum_value_list_2(self, p):
    """enum_value_list : nonempty_enum_value_list"""
    p[0] = p[1]

  def p_nonempty_enum_value_list_1(self, p):
    """nonempty_enum_value_list : enum_value"""
    p[0] = ast.EnumValueList(p[1])

  def p_nonempty_enum_value_list_2(self, p):
    """nonempty_enum_value_list : nonempty_enum_value_list COMMA enum_value"""
    p[0] = p[1]
    p[0].Append(p[3])

  def p_enum_value(self, p):
    """enum_value : attribute_section name_wrapped
                  | attribute_section name_wrapped EQUALS int
                  | attribute_section name_wrapped EQUALS identifier_wrapped"""
    p[0] = ast.EnumValue(
        p[2],
        p[1],
        p[4] if len(p) == 5 else None,
        filename=self.filename,
        lineno=p.lineno(2))

  def p_const(self, p):
    """const : attribute_section CONST typename name_wrapped EQUALS constant SEMI"""
    p[0] = ast.Const(p[4], p[1], p[3], p[6])

  def p_constant(self, p):
    """constant : literal
                | identifier_wrapped"""
    p[0] = p[1]

  def p_identifier_wrapped(self, p):
    """identifier_wrapped : identifier"""
    p[0] = ('IDENTIFIER', p[1])

  # TODO(vtl): Make this produce a "wrapped" identifier (probably as an
  # |ast.Identifier|, to be added) and get rid of identifier_wrapped.
  def p_identifier(self, p):
    """identifier : name_wrapped
                  | name_wrapped DOT identifier"""
    p[0] = ''.join(p[1:])

  # Allow 'feature' to be a name literal not just a keyword.
  def p_name_wrapped(self, p):
    """name_wrapped : NAME
                    | FEATURE"""
    p[0] = p[1]

  def p_literal(self, p):
    """literal : int
               | float
               | TRUE
               | FALSE
               | DEFAULT
               | STRING_LITERAL"""
    p[0] = p[1]

  def p_int(self, p):
    """int : int_const
           | PLUS int_const
           | MINUS int_const"""
    p[0] = ''.join(p[1:])

  def p_int_const(self, p):
    """int_const : INT_CONST_DEC
                 | INT_CONST_HEX"""
    p[0] = p[1]

  def p_float(self, p):
    """float : FLOAT_CONST
             | PLUS FLOAT_CONST
             | MINUS FLOAT_CONST"""
    p[0] = ''.join(p[1:])

  def p_error(self, e):
    if e is None:
      # Unexpected EOF.
      # TODO(vtl): Can we figure out what's missing?
      raise ParseError(self.filename, "Unexpected end of file")

    if e.value == 'feature':
      raise ParseError(self.filename,
                       "`feature` is reserved for a future mojom keyword",
                       lineno=e.lineno,
                       snippet=self._GetSnippet(e.lineno))

    raise ParseError(
        self.filename,
        "Unexpected %r:" % e.value,
        lineno=e.lineno,
        snippet=self._GetSnippet(e.lineno))

  def _GetSnippet(self, lineno):
    return self.source.split('\n')[lineno - 1]


def Parse(source, filename):
  """Parse source file to AST.

  Args:
    source: The source text as a str (Python 2 or 3) or unicode (Python 2).
    filename: The filename that |source| originates from.

  Returns:
    The AST as a mojom.parse.ast.Mojom object.
  """
  lexer = Lexer(filename)
  parser = Parser(lexer, source, filename)

  lex.lex(object=lexer)
  yacc.yacc(module=parser, debug=0, write_tables=0)

  tree = yacc.parse(source)
  return tree



# Copyright 2013 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import copy
from mojom.generate import module as mojom

# This module provides a mechanism for determining the packed order and offsets
# of a mojom.Struct.
#
# ps = pack.PackedStruct(struct)
# ps.packed_fields will access a list of PackedField objects, each of which
# will have an offset, a size and a bit (for mojom.BOOLs).

# Size of struct header in bytes: num_bytes [4B] + version [4B].
HEADER_SIZE = 8


class PackedField:
  kind_to_size = {
      mojom.BOOL: 1,
      mojom.INT8: 1,
      mojom.UINT8: 1,
      mojom.INT16: 2,
      mojom.UINT16: 2,
      mojom.INT32: 4,
      mojom.UINT32: 4,
      mojom.FLOAT: 4,
      mojom.HANDLE: 4,
      mojom.MSGPIPE: 4,
      mojom.SHAREDBUFFER: 4,
      mojom.PLATFORMHANDLE: 4,
      mojom.DCPIPE: 4,
      mojom.DPPIPE: 4,
      mojom.NULLABLE_HANDLE: 4,
      mojom.NULLABLE_MSGPIPE: 4,
      mojom.NULLABLE_SHAREDBUFFER: 4,
      mojom.NULLABLE_PLATFORMHANDLE: 4,
      mojom.NULLABLE_DCPIPE: 4,
      mojom.NULLABLE_DPPIPE: 4,
      mojom.INT64: 8,
      mojom.UINT64: 8,
      mojom.DOUBLE: 8,
      mojom.STRING: 8,
      mojom.NULLABLE_STRING: 8
  }

  @classmethod
  def GetSizeForKind(cls, kind):
    if isinstance(kind, (mojom.Array, mojom.Map, mojom.Struct, mojom.Interface,
                         mojom.AssociatedInterface, mojom.PendingRemote,
                         mojom.PendingAssociatedRemote)):
      return 8
    if isinstance(kind, mojom.Union):
      return 16
    if isinstance(kind, (mojom.InterfaceRequest, mojom.PendingReceiver)):
      kind = mojom.MSGPIPE
    if isinstance(
        kind,
        (mojom.AssociatedInterfaceRequest, mojom.PendingAssociatedReceiver)):
      return 4
    if isinstance(kind, mojom.Enum):
      # TODO(mpcomplete): what about big enums?
      return cls.kind_to_size[mojom.INT32]
    if not kind in cls.kind_to_size:
      raise Exception("Undefined type: %s. Did you forget to import the file "
                      "containing the definition?" % kind.spec)
    return cls.kind_to_size[kind]

  @classmethod
  def GetAlignmentForKind(cls, kind):
    if isinstance(kind, (mojom.Interface, mojom.AssociatedInterface,
                         mojom.PendingRemote, mojom.PendingAssociatedRemote)):
      return 4
    if isinstance(kind, mojom.Union):
      return 8
    return cls.GetSizeForKind(kind)

  def __init__(self,
               field,
               index,
               ordinal,
               original_field=None,
               sub_ordinal=None,
               linked_value_packed_field=None):
    """
    Args:
      field: the original field.
      index: the position of the original field in the struct.
      ordinal: the ordinal of the field for serialization.
      original_field: See below.
      sub_ordinal: See below.
      linked_value_packed_field: See below.

    original_field, sub_ordinal, and linked_value_packed_field are used to
    support nullable ValueKind fields. For legacy reasons, nullable ValueKind
    fields actually generate two PackedFields. This allows:

    - backwards compatibility prior to Mojo support for nullable ValueKinds.
    - correct packing of fields for the aforementioned backwards compatibility.

    When translating Fields to PackedFields, the original field is turned into
    two PackedFields: the first PackedField always has type mojom.BOOL, while
    the second PackedField has the non-nullable version of the field's kind.

    When constructing these PackedFields, original_field references the field
    as defined in the mojom; the name as defined in the mojom will be used for
    all layers above the wire/data layer.

    sub_ordinal is used to sort the two PackedFields correctly with respect to
    each other: the first mojom.BOOL field always has sub_ordinal 0, while the
    second field always has sub_ordinal 1.

    Finally, linked_value_packed_field is used by the serialization and
    deserialization helpers, which generally just iterate over a PackedStruct's
    PackedField's in ordinal order. This allows the helpers to easily reference
    any related PackedFields rather than having to lookup related PackedFields
    by index while iterating.
    """
    self.field = field
    self.index = index
    self.ordinal = ordinal
    self.original_field = original_field
    self.sub_ordinal = sub_ordinal
    self.linked_value_packed_field = linked_value_packed_field
    self.size = self.GetSizeForKind(self.field.kind)
    self.alignment = self.GetAlignmentForKind(self.field.kind)
    self.offset = None
    self.bit = None
    self.min_version = None


def GetPad(offset, alignment):
  """Returns the pad necessary to reserve space so that |offset + pad| equals to
  some multiple of |alignment|."""
  return (alignment - (offset % alignment)) % alignment


def GetFieldOffset(field, last_field):
  """Returns a 2-tuple of the field offset and bit (for BOOLs)."""
  if (field.field.kind == mojom.BOOL and last_field.field.kind == mojom.BOOL
      and last_field.bit < 7):
    return (last_field.offset, last_field.bit + 1)

  offset = last_field.offset + last_field.size
  pad = GetPad(offset, field.alignment)
  return (offset + pad, 0)


def GetPayloadSizeUpToField(field):
  """Returns the payload size (not including struct header) if |field| is the
  last field.
  """
  if not field:
    return 0
  offset = field.offset + field.size
  pad = GetPad(offset, 8)
  return offset + pad


def IsNullableValueKindPackedField(field):
  """Returns true if `field` is derived from a nullable ValueKind field.

  Nullable ValueKind fields often require special handling in the bindings due
  to the way the implementation is constrained for wire compatibility.
  """
  assert isinstance(field, PackedField)
  return field.sub_ordinal is not None


def IsPrimaryNullableValueKindPackedField(field):
  """Returns true if `field` is derived from a nullable ValueKind mojom field
  and is the "primary" field.

  The primary field is a bool PackedField that controls if the field should be
  considered as present or not; it will have a reference to the PackedField that
  holds the actual value representation if considered present.

  Bindings code that translates between the wire protocol and the higher layers
  can use this to simplify mapping multiple PackedFields to the single field
  that is logically exposed to bindings consumers.
  """
  assert isinstance(field, PackedField)
  return field.linked_value_packed_field is not None


class PackedStruct:
  def __init__(self, struct):
    self.struct = struct
    # |packed_fields| contains all the fields, in increasing offset order.
    self.packed_fields = []
    # |packed_fields_in_ordinal_order| refers to the same fields as
    # |packed_fields|, but in ordinal order.
    self.packed_fields_in_ordinal_order = []

    # No fields.
    if (len(struct.fields) == 0):
      return

    # Start by sorting by ordinal.
    src_fields = self.packed_fields_in_ordinal_order
    ordinal = 0
    for index, field in enumerate(struct.fields):
      if field.ordinal is not None:
        ordinal = field.ordinal
      # Nullable value types are a bit weird: they generate two PackedFields
      # despite being a single ValueKind. This is for wire compatibility to
      # ease the transition from legacy mojom syntax where nullable value types
      # were not supported.
      if isinstance(field.kind, mojom.ValueKind) and field.kind.is_nullable:
        # The suffixes intentionally use Unicode codepoints which are considered
        # valid C++/Java/JavaScript identifiers, yet are unlikely to be used in
        # actual user code.
        has_value_field = copy.copy(field)
        has_value_field.name = f'{field.mojom_name}_$flag'
        has_value_field.kind = mojom.BOOL

        value_field = copy.copy(field)
        value_field.name = f'{field.mojom_name}_$value'
        value_field.kind = field.kind.MakeUnnullableKind()

        value_packed_field = PackedField(value_field,
                                         index,
                                         ordinal,
                                         original_field=field,
                                         sub_ordinal=1,
                                         linked_value_packed_field=None)
        has_value_packed_field = PackedField(
            has_value_field,
            index,
            ordinal,
            original_field=field,
            sub_ordinal=0,
            linked_value_packed_field=value_packed_field)
        src_fields.append(has_value_packed_field)
        src_fields.append(value_packed_field)
      else:
        src_fields.append(PackedField(field, index, ordinal))
      ordinal += 1
    src_fields.sort(key=lambda field: (field.ordinal, field.sub_ordinal))

    # Set |min_version| for each field.
    next_min_version = 0
    for packed_field in src_fields:
      if packed_field.field.min_version is None:
        assert next_min_version == 0
      else:
        assert packed_field.field.min_version >= next_min_version
        next_min_version = packed_field.field.min_version
      packed_field.min_version = next_min_version

      if (packed_field.min_version != 0
          and mojom.IsReferenceKind(packed_field.field.kind)
          and not packed_field.field.kind.is_nullable):
        raise Exception(
            "Non-nullable reference fields are only allowed in version 0 of a "
            "struct. %s.%s is defined with [MinVersion=%d]." %
            (self.struct.name, packed_field.field.name,
             packed_field.min_version))

    src_field = src_fields[0]
    src_field.offset = 0
    src_field.bit = 0
    dst_fields = self.packed_fields
    dst_fields.append(src_field)

    # Then find first slot that each field will fit.
    for src_field in src_fields[1:]:
      last_field = dst_fields[0]
      for i in range(1, len(dst_fields)):
        next_field = dst_fields[i]
        offset, bit = GetFieldOffset(src_field, last_field)
        if offset + src_field.size <= next_field.offset:
          # Found hole.
          src_field.offset = offset
          src_field.bit = bit
          dst_fields.insert(i, src_field)
          break
        last_field = next_field
      if src_field.offset is None:
        # Add to end
        src_field.offset, src_field.bit = GetFieldOffset(src_field, last_field)
        dst_fields.append(src_field)


class ByteInfo:
  def __init__(self):
    self.is_padding = False
    self.packed_fields = []


def GetByteLayout(packed_struct):
  total_payload_size = GetPayloadSizeUpToField(
      packed_struct.packed_fields[-1] if packed_struct.packed_fields else None)
  byte_info = [ByteInfo() for i in range(total_payload_size)]

  limit_of_previous_field = 0
  for packed_field in packed_struct.packed_fields:
    for i in range(limit_of_previous_field, packed_field.offset):
      byte_info[i].is_padding = True
    byte_info[packed_field.offset].packed_fields.append(packed_field)
    limit_of_previous_field = packed_field.offset + packed_field.size

  for i in range(limit_of_previous_field, len(byte_info)):
    byte_info[i].is_padding = True

  for byte in byte_info:
    # A given byte cannot both be padding and have a fields packed into it.
    assert not (byte.is_padding and byte.packed_fields)

  return byte_info


class VersionInfo:
  def __init__(self, version, num_fields, num_packed_fields, num_bytes):
    self.version = version
    self.num_fields = num_fields
    self.num_packed_fields = num_packed_fields
    self.num_bytes = num_bytes


def GetVersionInfo(packed_struct):
  """Get version information for a struct.

  Args:
    packed_struct: A PackedStruct instance.

  Returns:
    A non-empty list of VersionInfo instances, sorted by version in increasing
    order.
    Note: The version numbers may not be consecutive.
  """
  versions = []
  last_version = 0
  last_num_fields = 0
  last_num_packed_fields = 0
  last_payload_size = 0

  for packed_field in packed_struct.packed_fields_in_ordinal_order:
    if packed_field.min_version != last_version:
      versions.append(
          VersionInfo(last_version, last_num_fields, last_num_packed_fields,
                      last_payload_size + HEADER_SIZE))
      last_version = packed_field.min_version

    # Nullable numeric fields (e.g. `int32?`) expand to two packed fields, so to
    # avoid double-counting, only increment if the field is:
    # - not used for representing a nullable value kind field, or
    # - the primary field representing the nullable value kind field.
    last_num_fields += 1 if (
        not IsNullableValueKindPackedField(packed_field)
        or IsPrimaryNullableValueKindPackedField(packed_field)) else 0

    last_num_packed_fields += 1

    # The fields are iterated in ordinal order here. However, the size of a
    # version is determined by the last field of that version in pack order,
    # instead of ordinal order. Therefore, we need to calculate the max value.
    last_payload_size = max(GetPayloadSizeUpToField(packed_field),
                            last_payload_size)

  assert len(
      versions) == 0 or last_num_packed_fields != versions[-1].num_packed_fields
  versions.append(
      VersionInfo(last_version, last_num_fields, last_num_packed_fields,
                  last_payload_size + HEADER_SIZE))
  return versions



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# smoothing.py - Wrapper for cv2 smoothing functions to enable duck-typing

import cv2


# @brief Wrapper for cv2 smoothing functions so that they can be duck-typed
class Smoothing(object):
    def __init__(self):
        pass

    def smoothing(self, src):
        raise NotImplementedError


class MedianBlur(Smoothing):
    def __init__(self, ksize):
        self.ksize = ksize

    def smoothing(self, src):
        return cv2.medianBlur(src.astype('float32'), self.ksize).astype('float64')



#!/usr/bin/env python
# Copyright 2013 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""The frontend for the Mojo bindings system."""

from __future__ import print_function

import argparse

import hashlib
import importlib
import json
import os
import pprint
import re
import struct
import sys

# Disable lint check for finding modules:
# pylint: disable=F0401

def _GetDirAbove(dirname):
  """Returns the directory "above" this file containing |dirname| (which must
  also be "above" this file)."""
  path = os.path.abspath(__file__)
  while True:
    path, tail = os.path.split(path)
    assert tail
    if tail == dirname:
      return path


sys.path.insert(
    0,
    os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "mojom"))

from mojom.error import Error
import mojom.fileutil as fileutil
from mojom.generate.module import Module
from mojom.generate import template_expander
from mojom.generate import translate
from mojom.generate.generator import WriteFile

sys.path.append(
    os.path.join(_GetDirAbove("mojo"), "tools", "diagnosis"))
import crbug_1001171


_BUILTIN_GENERATORS = {
    "c++": "mojom_cpp_generator",
    "javascript": "mojom_js_generator",
    "java": "mojom_java_generator",
    "mojolpm": "mojom_mojolpm_generator",
    "typescript": "mojom_ts_generator",
}

_BUILTIN_CHECKS = {
    "attributes": "mojom_attributes_check",
    "definitions": "mojom_definitions_check",
    "features": "mojom_interface_feature_check",
    "restrictions": "mojom_restrictions_check",
}


def LoadGenerators(generators_string):
  if not generators_string:
    return {}  # No generators.

  generators = {}
  for generator_name in [s.strip() for s in generators_string.split(",")]:
    language = generator_name.lower()
    if language not in _BUILTIN_GENERATORS:
      print("Unknown generator name %s" % generator_name)
      sys.exit(1)
    generator_module = importlib.import_module(
        "generators.%s" % _BUILTIN_GENERATORS[language])
    generators[language] = generator_module
  return generators


def LoadChecks(checks_string):
  if not checks_string:
    return {}  # No checks.

  checks = {}
  for check_name in [s.strip() for s in checks_string.split(",")]:
    check = check_name.lower()
    if check not in _BUILTIN_CHECKS:
      print("Unknown check name %s" % check_name)
      sys.exit(1)
    check_module = importlib.import_module("checks.%s" % _BUILTIN_CHECKS[check])
    checks[check] = check_module
  return checks


def MakeImportStackMessage(imported_filename_stack):
  """Make a (human-readable) message listing a chain of imports. (Returned
  string begins with a newline (if nonempty) and does not end with one.)"""
  return ''.join(
      reversed(["\n  %s was imported by %s" % (a, b) for (a, b) in \
                    zip(imported_filename_stack[1:], imported_filename_stack)]))


class RelativePath:
  """Represents a path relative to the source tree or generated output dir."""

  def __init__(self, path, source_root, output_dir):
    self.path = path
    if path.startswith(source_root):
      self.root = source_root
    elif path.startswith(output_dir):
      self.root = output_dir
    else:
      raise Exception("Invalid input path %s" % path)

  def relative_path(self):
    return os.path.relpath(
        os.path.abspath(self.path), os.path.abspath(self.root))


def _GetModulePath(path, output_dir):
  return os.path.join(output_dir, path.relative_path() + '-module')


def ScrambleMethodOrdinals(interfaces, salt):
  already_generated = set()
  for interface in interfaces:
    i = 0
    already_generated.clear()
    for method in interface.methods:
      if method.explicit_ordinal is not None:
        continue
      while True:
        i = i + 1
        if i == 1000000:
          raise Exception("Could not generate %d method ordinals for %s" %
              (len(interface.methods), interface.mojom_name))
        # Generate a scrambled method.ordinal value. The algorithm doesn't have
        # to be very strong, cryptographically. It just needs to be non-trivial
        # to guess the results without the secret salt, in order to make it
        # harder for a compromised process to send fake Mojo messages.
        sha256 = hashlib.sha256(salt)
        sha256.update(interface.mojom_name.encode('utf-8'))
        sha256.update(str(i).encode('utf-8'))
        # Take the first 4 bytes as a little-endian uint32.
        ordinal = struct.unpack('<L', sha256.digest()[:4])[0]
        # Trim to 31 bits, so it always fits into a Java (signed) int.
        ordinal = ordinal & 0x7fffffff
        if ordinal in already_generated:
          continue
        already_generated.add(ordinal)
        method.ordinal = ordinal
        method.ordinal_comment = (
            'The %s value is based on sha256(salt + "%s%d").' %
            (ordinal, interface.mojom_name, i))
        break


def ReadFileContents(filename):
  with open(filename, 'rb') as f:
    return f.read()


class MojomProcessor:
  """Takes parsed mojom modules and generates language bindings from them.

  Attributes:
    _processed_files: {Dict[str, mojom.generate.module.Module]} Mapping from
        relative mojom filename paths to the module AST for that mojom file.
  """
  def __init__(self, should_generate):
    self._should_generate = should_generate
    self._processed_files = {}
    self._typemap = {}

  def LoadTypemaps(self, typemaps):
    # Support some very simple single-line comments in typemap JSON.
    comment_expr = r"^\s*//.*$"
    def no_comments(line):
      return not re.match(comment_expr, line)
    for filename in typemaps:
      with open(filename) as f:
        typemaps = json.loads("".join(filter(no_comments, f.readlines())))
        for language, typemap in typemaps.items():
          language_map = self._typemap.get(language, {})
          language_map.update(typemap)
          self._typemap[language] = language_map
    if 'c++' in self._typemap:
      self._typemap['mojolpm'] = self._typemap['c++']

  def _GenerateModule(self, args, remaining_args, check_modules,
                      generator_modules, rel_filename, imported_filename_stack):
    # Return the already-generated module.
    if rel_filename.path in self._processed_files:
      return self._processed_files[rel_filename.path]

    if rel_filename.path in imported_filename_stack:
      print("%s: Error: Circular dependency" % rel_filename.path + \
          MakeImportStackMessage(imported_filename_stack + [rel_filename.path]))
      sys.exit(1)

    module_path = _GetModulePath(rel_filename, args.output_dir)
    with open(module_path, 'rb') as f:
      module = Module.Load(f)

    if args.scrambled_message_id_salt_paths:
      salt = b''.join(
          map(ReadFileContents, args.scrambled_message_id_salt_paths))
      ScrambleMethodOrdinals(module.interfaces, salt)

    if self._should_generate(rel_filename.path):
      # Run checks on module first.
      for check_module in check_modules.values():
        checker = check_module.Check(module)
        checker.CheckModule()
      # Then run generation.
      for language, generator_module in generator_modules.items():
        generator = generator_module.Generator(
            module, args.output_dir, typemap=self._typemap.get(language, {}),
            variant=args.variant, bytecode_path=args.bytecode_path,
            for_blink=args.for_blink,
            js_generate_struct_deserializers=\
                    args.js_generate_struct_deserializers,
            export_attribute=args.export_attribute,
            export_header=args.export_header,
            generate_non_variant_code=args.generate_non_variant_code,
            support_lazy_serialization=args.support_lazy_serialization,
            disallow_native_types=args.disallow_native_types,
            disallow_interfaces=args.disallow_interfaces,
            generate_message_ids=args.generate_message_ids,
            generate_fuzzing=args.generate_fuzzing,
            enable_kythe_annotations=args.enable_kythe_annotations,
            extra_cpp_template_paths=args.extra_cpp_template_paths,
            generate_extra_cpp_only=args.generate_extra_cpp_only)
        filtered_args = []
        if hasattr(generator_module, 'GENERATOR_PREFIX'):
          prefix = '--' + generator_module.GENERATOR_PREFIX + '_'
          filtered_args = [arg for arg in remaining_args
                           if arg.startswith(prefix)]
        generator.GenerateFiles(filtered_args)

    # Save result.
    self._processed_files[rel_filename.path] = module
    return module


def _Generate(args, remaining_args):
  if args.variant == "none":
    args.variant = None

  for idx, import_dir in enumerate(args.import_directories):
    tokens = import_dir.split(":")
    if len(tokens) >= 2:
      args.import_directories[idx] = RelativePath(tokens[0], tokens[1],
                                                  args.output_dir)
    else:
      args.import_directories[idx] = RelativePath(tokens[0], args.depth,
                                                  args.output_dir)
  generator_modules = LoadGenerators(args.generators_string)
  check_modules = LoadChecks(args.checks_string)

  fileutil.EnsureDirectoryExists(args.output_dir)

  processor = MojomProcessor(lambda filename: filename in args.filename)
  processor.LoadTypemaps(set(args.typemaps))

  if args.filelist:
    with open(args.filelist) as f:
      args.filename.extend(f.read().split())

  for filename in args.filename:
    processor._GenerateModule(
        args, remaining_args, check_modules, generator_modules,
        RelativePath(filename, args.depth, args.output_dir), [])

  return 0


def _Precompile(args, _):
  generator_modules = LoadGenerators(",".join(_BUILTIN_GENERATORS.keys()))

  template_expander.PrecompileTemplates(generator_modules, args.output_dir)
  return 0


def main():
  parser = argparse.ArgumentParser(
      description="Generate bindings from mojom files.")
  parser.add_argument("--use_bundled_pylibs", action="store_true",
                      help="use Python modules bundled in the SDK")
  parser.add_argument(
      "-o",
      "--output_dir",
      dest="output_dir",
      default=".",
      help="output directory for generated files")

  subparsers = parser.add_subparsers()

  generate_parser = subparsers.add_parser(
      "generate", description="Generate bindings from mojom files.")
  generate_parser.add_argument("filename", nargs="*",
                               help="mojom input file")
  generate_parser.add_argument("--filelist", help="mojom input file list")
  generate_parser.add_argument("-d", "--depth", dest="depth", default=".",
                               help="depth from source root")
  generate_parser.add_argument("-g",
                               "--generators",
                               dest="generators_string",
                               metavar="GENERATORS",
                               default="c++,javascript,java,mojolpm",
                               help="comma-separated list of generators")
  generate_parser.add_argument("-c",
                               "--checks",
                               dest="checks_string",
                               metavar="CHECKS",
                               default=",".join(_BUILTIN_CHECKS.keys()),
                               help="comma-separated list of checks")
  generate_parser.add_argument(
      "--gen_dir", dest="gen_directories", action="append", metavar="directory",
      default=[], help="add a directory to be searched for the syntax trees.")
  generate_parser.add_argument(
      "-I", dest="import_directories", action="append", metavar="directory",
      default=[],
      help="add a directory to be searched for import files. The depth from "
           "source root can be specified for each import by appending it after "
           "a colon")
  generate_parser.add_argument("--typemap", action="append", metavar="TYPEMAP",
                               default=[], dest="typemaps",
                               help="apply TYPEMAP to generated output")
  generate_parser.add_argument("--variant", dest="variant", default=None,
                               help="output a named variant of the bindings")
  generate_parser.add_argument(
      "--bytecode_path", required=True, help=(
          "the path from which to load template bytecode; to generate template "
          "bytecode, run %s precompile BYTECODE_PATH" % os.path.basename(
              sys.argv[0])))
  generate_parser.add_argument("--for_blink", action="store_true",
                               help="Use WTF types as generated types for mojo "
                               "string/array/map.")
  generate_parser.add_argument(
      "--js_generate_struct_deserializers", action="store_true",
      help="Generate javascript deserialize methods for structs in "
      "mojom-lite.js file")
  generate_parser.add_argument(
      "--export_attribute", default="",
      help="Optional attribute to specify on class declaration to export it "
      "for the component build.")
  generate_parser.add_argument(
      "--export_header", default="",
      help="Optional header to include in the generated headers to support the "
      "component build.")
  generate_parser.add_argument(
      "--generate_non_variant_code", action="store_true",
      help="Generate code that is shared by different variants.")
  generate_parser.add_argument(
      "--scrambled_message_id_salt_path",
      dest="scrambled_message_id_salt_paths",
      help="If non-empty, the path to a file whose contents should be used as"
      "a salt for generating scrambled message IDs. If this switch is specified"
      "more than once, the contents of all salt files are concatenated to form"
      "the salt value.", default=[], action="append")
  generate_parser.add_argument(
      "--support_lazy_serialization",
      help="If set, generated bindings will serialize lazily when possible.",
      action="store_true")
  generate_parser.add_argument(
      "--extra_cpp_template_paths",
      dest="extra_cpp_template_paths",
      action="append",
      metavar="path_to_template",
      default=[],
      help="Provide a path to a new template (.tmpl) that is used to generate "
      "additional C++ source/header files ")
  generate_parser.add_argument(
      "--generate_extra_cpp_only",
      help="If set and extra_cpp_template_paths provided, will only generate"
      "extra_cpp_template related C++ bindings",
      action="store_true")
  generate_parser.add_argument(
      "--disallow_native_types",
      help="Disallows the [Native] attribute to be specified on structs or "
      "enums within the mojom file.", action="store_true")
  generate_parser.add_argument(
      "--disallow_interfaces",
      help="Disallows interface definitions within the mojom file. It is an "
      "error to specify this flag when processing a mojom file which defines "
      "any interface.", action="store_true")
  generate_parser.add_argument(
      "--generate_message_ids",
      help="Generates only the message IDs header for C++ bindings. Note that "
      "this flag only matters if --generate_non_variant_code is also "
      "specified.", action="store_true")
  generate_parser.add_argument(
      "--generate_fuzzing",
      action="store_true",
      help="Generates additional bindings for fuzzing in JS.")
  generate_parser.add_argument(
      "--enable_kythe_annotations",
      action="store_true",
      help="Adds annotations for kythe metadata generation.")

  generate_parser.set_defaults(func=_Generate)

  precompile_parser = subparsers.add_parser("precompile",
      description="Precompile templates for the mojom bindings generator.")
  precompile_parser.set_defaults(func=_Precompile)

  args, remaining_args = parser.parse_known_args()
  return args.func(args, remaining_args)


if __name__ == "__main__":
  with crbug_1001171.DumpStateOnLookupError():
    ret = main()
    # Exit without running GC, which can save multiple seconds due to the large
    # number of object created. But flush is necessary as os._exit doesn't do
    # that.
    sys.stdout.flush()
    sys.stderr.flush()
    os._exit(ret)



# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import json

from mojom_parser_test_case import MojomParserTestCase


class MojomParserTest(MojomParserTestCase):
  """Tests covering the behavior defined by the main mojom_parser.py script.
  This includes behavior around input and output path manipulation, dependency
  resolution, and module serialization and deserialization."""

  def testBasicParse(self):
    """Basic test to verify that we can parse a mojom file and get a module."""
    mojom = 'foo/bar.mojom'
    self.WriteFile(
        mojom, """\
        module test;
        enum TestEnum { kFoo };
        """)
    self.ParseMojoms([mojom])

    m = self.LoadModule(mojom)
    self.assertEqual('foo/bar.mojom', m.path)
    self.assertEqual('test', m.mojom_namespace)
    self.assertEqual(1, len(m.enums))

  def testBasicParseWithAbsolutePaths(self):
    """Verifies that we can parse a mojom file given an absolute path input."""
    mojom = 'foo/bar.mojom'
    self.WriteFile(
        mojom, """\
        module test;
        enum TestEnum { kFoo };
        """)
    self.ParseMojoms([self.GetPath(mojom)])

    m = self.LoadModule(mojom)
    self.assertEqual('foo/bar.mojom', m.path)
    self.assertEqual('test', m.mojom_namespace)
    self.assertEqual(1, len(m.enums))

  def testImport(self):
    """Verify imports within the same set of mojom inputs."""
    a = 'a.mojom'
    b = 'b.mojom'
    self.WriteFile(
        a, """\
        module a;
        import "b.mojom";
        struct Foo { b.Bar bar; };""")
    self.WriteFile(b, """\
        module b;
        struct Bar {};""")
    self.ParseMojoms([a, b])

    ma = self.LoadModule(a)
    mb = self.LoadModule(b)
    self.assertEqual('a.mojom', ma.path)
    self.assertEqual('b.mojom', mb.path)
    self.assertEqual(1, len(ma.imports))
    self.assertEqual(mb, ma.imports[0])

  def testPreProcessedImport(self):
    """Verify imports processed by a previous parser execution can be loaded
    properly when parsing a dependent mojom."""
    a = 'a.mojom'
    self.WriteFile(a, """\
        module a;
        struct Bar {};""")
    self.ParseMojoms([a])

    b = 'b.mojom'
    self.WriteFile(
        b, """\
        module b;
        import "a.mojom";
        struct Foo { a.Bar bar; };""")
    self.ParseMojoms([b])

  def testMissingImport(self):
    """Verify that an import fails if the imported mojom does not exist."""
    a = 'a.mojom'
    self.WriteFile(
        a, """\
        module a;
        import "non-existent.mojom";
        struct Bar {};""")
    with self.assertRaisesRegexp(ValueError, "does not exist"):
      self.ParseMojoms([a])

  def testUnparsedImport(self):
    """Verify that an import fails if the imported mojom is not in the set of
    mojoms provided to the parser on this execution AND there is no pre-existing
    parsed output module already on disk for it."""
    a = 'a.mojom'
    b = 'b.mojom'
    self.WriteFile(a, """\
        module a;
        struct Bar {};""")
    self.WriteFile(
        b, """\
        module b;
        import "a.mojom";
        struct Foo { a.Bar bar; };""")

    # a.mojom has not been parsed yet, so its import will fail when processing
    # b.mojom here.
    with self.assertRaisesRegexp(ValueError, "does not exist"):
      self.ParseMojoms([b])

  def testCheckImportsBasic(self):
    """Verify that the parser can handle --check-imports with a valid set of
    inputs, including support for transitive dependency resolution."""
    a = 'a.mojom'
    a_metadata = 'out/a.build_metadata'
    b = 'b.mojom'
    b_metadata = 'out/b.build_metadata'
    c = 'c.mojom'
    c_metadata = 'out/c.build_metadata'
    self.WriteFile(a_metadata,
                   json.dumps({
                       "sources": [self.GetPath(a)],
                       "deps": []
                   }))
    self.WriteFile(
        b_metadata,
        json.dumps({
            "sources": [self.GetPath(b)],
            "deps": [self.GetPath(a_metadata)]
        }))
    self.WriteFile(
        c_metadata,
        json.dumps({
            "sources": [self.GetPath(c)],
            "deps": [self.GetPath(b_metadata)]
        }))
    self.WriteFile(a, """\
        module a;
        struct Bar {};""")
    self.WriteFile(
        b, """\
        module b;
        import "a.mojom";
        struct Foo { a.Bar bar; };""")
    self.WriteFile(
        c, """\
        module c;
        import "a.mojom";
        import "b.mojom";
        struct Baz { b.Foo foo; };""")
    self.ParseMojoms([a], metadata=a_metadata)
    self.ParseMojoms([b], metadata=b_metadata)
    self.ParseMojoms([c], metadata=c_metadata)

  def testCheckImportsMissing(self):
    """Verify that the parser rejects valid input mojoms when imports don't
    agree with build metadata given via --check-imports."""
    a = 'a.mojom'
    a_metadata = 'out/a.build_metadata'
    b = 'b.mojom'
    b_metadata = 'out/b.build_metadata'
    self.WriteFile(a_metadata,
                   json.dumps({
                       "sources": [self.GetPath(a)],
                       "deps": []
                   }))
    self.WriteFile(b_metadata,
                   json.dumps({
                       "sources": [self.GetPath(b)],
                       "deps": []
                   }))
    self.WriteFile(a, """\
        module a;
        struct Bar {};""")
    self.WriteFile(
        b, """\
        module b;
        import "a.mojom";
        struct Foo { a.Bar bar; };""")

    self.ParseMojoms([a], metadata=a_metadata)
    with self.assertRaisesRegexp(ValueError, "not allowed by build"):
      self.ParseMojoms([b], metadata=b_metadata)



# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from mojom_parser_test_case import MojomParserTestCase

from mojom.generate import module


class StableAttributeTest(MojomParserTestCase):
  """Tests covering usage of the [Stable] attribute."""

  def testStableAttributeTagging(self):
    """Verify that we recognize the [Stable] attribute on relevant definitions
    and the resulting parser outputs are tagged accordingly."""
    mojom = 'test.mojom'
    self.WriteFile(
        mojom, """\
        [Stable] enum TestEnum { kFoo };
        enum UnstableEnum { kBar };
        [Stable] struct TestStruct { TestEnum a; };
        struct UnstableStruct { UnstableEnum a; };
        [Stable] union TestUnion { TestEnum a; TestStruct b; };
        union UnstableUnion { UnstableEnum a; UnstableStruct b; };
        [Stable] interface TestInterface { Foo@0(TestUnion x) => (); };
        interface UnstableInterface { Foo(UnstableUnion x) => (); };
        """)
    self.ParseMojoms([mojom])

    m = self.LoadModule(mojom)
    self.assertEqual(2, len(m.enums))
    self.assertTrue(m.enums[0].stable)
    self.assertFalse(m.enums[1].stable)
    self.assertEqual(2, len(m.structs))
    self.assertTrue(m.structs[0].stable)
    self.assertFalse(m.structs[1].stable)
    self.assertEqual(2, len(m.unions))
    self.assertTrue(m.unions[0].stable)
    self.assertFalse(m.unions[1].stable)
    self.assertEqual(2, len(m.interfaces))
    self.assertTrue(m.interfaces[0].stable)
    self.assertFalse(m.interfaces[1].stable)

  def testStableStruct(self):
    """A [Stable] struct is valid if all its fields are also stable."""
    self.ExtractTypes('[Stable] struct S {};')
    self.ExtractTypes('[Stable] struct S { int32 x; bool b; };')
    self.ExtractTypes('[Stable] enum E { A }; [Stable] struct S { E e; };')
    self.ExtractTypes('[Stable] struct S {}; [Stable] struct T { S s; };')
    self.ExtractTypes(
        '[Stable] struct S {}; [Stable] struct T { array<S> ss; };')
    self.ExtractTypes(
        '[Stable] interface F {}; [Stable] struct T { pending_remote<F> f; };')

    with self.assertRaisesRegexp(Exception, 'because it depends on E'):
      self.ExtractTypes('enum E { A }; [Stable] struct S { E e; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on X'):
      self.ExtractTypes('struct X {}; [Stable] struct S { X x; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] struct S { array<T> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] struct S { map<int32, T> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] struct S { map<T, int32> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on F'):
      self.ExtractTypes(
          'interface F {}; [Stable] struct S { pending_remote<F> f; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on F'):
      self.ExtractTypes(
          'interface F {}; [Stable] struct S { pending_receiver<F> f; };')

  def testStableUnion(self):
    """A [Stable] union is valid if all its fields' types are also stable."""
    self.ExtractTypes('[Stable] union U {};')
    self.ExtractTypes('[Stable] union U { int32 x; bool b; };')
    self.ExtractTypes('[Stable] enum E { A }; [Stable] union U { E e; };')
    self.ExtractTypes('[Stable] struct S {}; [Stable] union U { S s; };')
    self.ExtractTypes(
        '[Stable] struct S {}; [Stable] union U { array<S> ss; };')
    self.ExtractTypes(
        '[Stable] interface F {}; [Stable] union U { pending_remote<F> f; };')

    with self.assertRaisesRegexp(Exception, 'because it depends on E'):
      self.ExtractTypes('enum E { A }; [Stable] union U { E e; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on X'):
      self.ExtractTypes('struct X {}; [Stable] union U { X x; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] union U { array<T> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] union U { map<int32, T> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on T'):
      self.ExtractTypes('struct T {}; [Stable] union U { map<T, int32> xs; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on F'):
      self.ExtractTypes(
          'interface F {}; [Stable] union U { pending_remote<F> f; };')
    with self.assertRaisesRegexp(Exception, 'because it depends on F'):
      self.ExtractTypes(
          'interface F {}; [Stable] union U { pending_receiver<F> f; };')

  def testStableInterface(self):
    """A [Stable] interface is valid if all its methods' parameter types are
    stable, including response parameters where applicable."""
    self.ExtractTypes('[Stable] interface F {};')
    self.ExtractTypes('[Stable] interface F { A@0(int32 x); };')
    self.ExtractTypes('[Stable] interface F { A@0(int32 x) => (bool b); };')
    self.ExtractTypes("""\
        [Stable] enum E { A, B, C };
        [Stable] struct S {};
        [Stable] interface F { A@0(E e, S s) => (bool b, array<S> s); };
        """)

    with self.assertRaisesRegexp(Exception, 'because it depends on E'):
      self.ExtractTypes(
          'enum E { A, B, C }; [Stable] interface F { A@0(E e); };')
    with self.assertRaisesRegexp(Exception, 'because it depends on E'):
      self.ExtractTypes(
          'enum E { A, B, C }; [Stable] interface F { A@0(int32 x) => (E e); };'
      )
    with self.assertRaisesRegexp(Exception, 'because it depends on S'):
      self.ExtractTypes(
          'struct S {}; [Stable] interface F { A@0(int32 x) => (S s); };')
    with self.assertRaisesRegexp(Exception, 'because it depends on S'):
      self.ExtractTypes(
          'struct S {}; [Stable] interface F { A@0(S s) => (bool b); };')

    with self.assertRaisesRegexp(Exception, 'explicit method ordinals'):
      self.ExtractTypes('[Stable] interface F { A() => (); };')



#!/usr/bin/env python3
# Copyright 2020 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Parses mojom IDL files.

This script parses one or more input mojom files and produces corresponding
module files fully describing the definitions contained within each mojom. The
module data is pickled and can be easily consumed by other tools to, e.g.,
generate usable language bindings.
"""

import argparse
import builtins
import codecs
import errno
import json
import logging
import multiprocessing
import os
import os.path
import sys
import traceback
from collections import defaultdict

from mojom.generate import module
from mojom.generate import translate
from mojom.parse import parser
from mojom.parse import conditional_features


# Disable this for easier debugging.
_ENABLE_MULTIPROCESSING = True

# https://docs.python.org/3/library/multiprocessing.html#:~:text=bpo-33725
if __name__ == '__main__' and sys.platform == 'darwin':
  multiprocessing.set_start_method('fork')
_MULTIPROCESSING_USES_FORK = multiprocessing.get_start_method() == 'fork'


def _ResolveRelativeImportPath(path, roots):
  """Attempts to resolve a relative import path against a set of possible roots.

  Args:
    path: The relative import path to resolve.
    roots: A list of absolute paths which will be checked in descending length
        order for a match against path.

  Returns:
    A normalized absolute path combining one of the roots with the input path if
    and only if such a file exists.

  Raises:
    ValueError: The path could not be resolved against any of the given roots.
  """
  for root in reversed(sorted(roots, key=len)):
    abs_path = os.path.join(root, path)
    if os.path.isfile(abs_path):
      return os.path.normcase(os.path.normpath(abs_path))

  raise ValueError('"%s" does not exist in any of %s' % (path, roots))


def RebaseAbsolutePath(path, roots):
  """Rewrites an absolute file path as relative to an absolute directory path in
  roots.

  Args:
    path: The absolute path of an existing file.
    roots: A list of absolute directory paths. The given path argument must fall
        within one of these directories.

  Returns:
    A path equivalent to the input path, but relative to one of the provided
    roots. If the input path falls within multiple roots, the longest root is
    chosen (and thus the shortest relative path is returned).

    Paths returned by this method always use forward slashes as a separator to
    mirror mojom import syntax.

  Raises:
    ValueError if the given path does not fall within any of the listed roots.
  """
  assert os.path.isabs(path)
  assert os.path.isfile(path)
  assert all(map(os.path.isabs, roots))

  sorted_roots = list(reversed(sorted(roots, key=len)))

  def try_rebase_path(path, root):
    head, rebased_path = os.path.split(path)
    while head != root:
      head, tail = os.path.split(head)
      if not tail:
        return None
      rebased_path = os.path.join(tail, rebased_path)
    return rebased_path

  for root in sorted_roots:
    relative_path = try_rebase_path(path, root)
    if relative_path:
      # TODO(crbug.com/953884): Use pathlib for this kind of thing once we're
      # fully migrated to Python 3.
      return relative_path.replace('\\', '/')

  raise ValueError('%s does not fall within any of %s' % (path, sorted_roots))


def _GetModuleFilename(mojom_filename):
  return mojom_filename + '-module'


def _EnsureInputLoaded(mojom_abspath, module_path, abs_paths, asts,
                       dependencies, loaded_modules, module_metadata):
  """Recursively ensures that a module and its dependencies are loaded.

  Args:
    mojom_abspath: An absolute file path pointing to a mojom file to load.
    module_path: The relative path used to identify mojom_abspath.
    abs_paths: A mapping from module paths to absolute file paths for all
        inputs given to this execution of the script.
    asts: A map from each input mojom's absolute path to its parsed AST.
    dependencies: A mapping of which input mojoms depend on each other, indexed
        by absolute file path.
    loaded_modules: A mapping of all modules loaded so far, including non-input
        modules that were pulled in as transitive dependencies of the inputs.
    module_metadata: Metadata to be attached to every module loaded by this
        helper.

  Returns:
    None

    On return, loaded_modules will be populated with the loaded input mojom's
    Module as well as the Modules of all of its transitive dependencies."""

  if mojom_abspath in loaded_modules:
    # Already done.
    return

  for dep_abspath, dep_path in sorted(dependencies[mojom_abspath]):
    if dep_abspath not in loaded_modules:
      _EnsureInputLoaded(dep_abspath, dep_path, abs_paths, asts, dependencies,
                         loaded_modules, module_metadata)

  imports = {}
  for imp in asts[mojom_abspath].import_list:
    path = imp.import_filename
    imports[path] = loaded_modules[abs_paths[path]]
  loaded_modules[mojom_abspath] = translate.OrderedModule(
      asts[mojom_abspath], module_path, imports)
  loaded_modules[mojom_abspath].metadata = dict(module_metadata)


def _CollectAllowedImportsFromBuildMetadata(build_metadata_filename):
  allowed_imports = set()
  processed_deps = set()

  def collect(metadata_filename):
    processed_deps.add(metadata_filename)

    # Paths in the metadata file are relative to the metadata file's dir.
    metadata_dir = os.path.abspath(os.path.dirname(metadata_filename))

    def to_abs(s):
      return os.path.normpath(os.path.join(metadata_dir, s))

    with open(metadata_filename) as f:
      metadata = json.load(f)
      allowed_imports.update(
          [os.path.normcase(to_abs(s)) for s in metadata['sources']])
      for dep_metadata in metadata['deps']:
        dep_metadata = to_abs(dep_metadata)
        if dep_metadata not in processed_deps:
          collect(dep_metadata)

  collect(build_metadata_filename)
  return allowed_imports


# multiprocessing helper.
def _ParseAstHelper(mojom_abspath, enabled_features):
  with codecs.open(mojom_abspath, encoding='utf-8') as f:
    ast = parser.Parse(f.read(), mojom_abspath)
    conditional_features.RemoveDisabledDefinitions(ast, enabled_features)
    return mojom_abspath, ast


# multiprocessing helper.
def _SerializeHelper(mojom_abspath, mojom_path):
  module_path = os.path.join(_SerializeHelper.output_root_path,
                             _GetModuleFilename(mojom_path))
  module_dir = os.path.dirname(module_path)
  if not os.path.exists(module_dir):
    try:
      # Python 2 doesn't support exist_ok on makedirs(), so we just ignore
      # that failure if it happens. It's possible during build due to races
      # among build steps with module outputs in the same directory.
      os.makedirs(module_dir)
    except OSError as e:
      if e.errno != errno.EEXIST:
        raise
  with open(module_path, 'wb') as f:
    _SerializeHelper.loaded_modules[mojom_abspath].Dump(f)


class _ExceptionWrapper:
  def __init__(self):
    # Do not capture exception object to ensure pickling works.
    self.formatted_trace = traceback.format_exc()


class _FuncWrapper:
  """Marshals exceptions and spreads args."""

  def __init__(self, func):
    self._func = func

  def __call__(self, args):
    # multiprocessing does not gracefully handle excptions.
    # https://crbug.com/1219044
    try:
      return self._func(*args)
    except:  # pylint: disable=bare-except
      return _ExceptionWrapper()


def _Shard(target_func, arg_list, processes=None):
  arg_list = list(arg_list)
  if processes is None:
    processes = multiprocessing.cpu_count()
  # Seems optimal to have each process perform at least 2 tasks.
  processes = min(processes, len(arg_list) // 2)

  if sys.platform == 'win32':
    # TODO(crbug.com/1190269) - we can't use more than 56
    # cores on Windows or Python3 may hang.
    processes = min(processes, 56)

  # Don't spin up processes unless there is enough work to merit doing so.
  if not _ENABLE_MULTIPROCESSING or processes < 2:
    for arg_tuple in arg_list:
      yield target_func(*arg_tuple)
    return

  pool = multiprocessing.Pool(processes=processes)
  try:
    wrapped_func = _FuncWrapper(target_func)
    for result in pool.imap_unordered(wrapped_func, arg_list):
      if isinstance(result, _ExceptionWrapper):
        sys.stderr.write(result.formatted_trace)
        sys.exit(1)
      yield result
  finally:
    pool.close()
    pool.join()  # Needed on Windows to avoid WindowsError during terminate.
    pool.terminate()


def _ParseMojoms(mojom_files,
                 input_root_paths,
                 output_root_path,
                 module_root_paths,
                 enabled_features,
                 module_metadata,
                 allowed_imports=None):
  """Parses a set of mojom files and produces serialized module outputs.

  Args:
    mojom_files: A list of mojom files to process. Paths must be absolute paths
        which fall within one of the input or output root paths.
    input_root_paths: A list of absolute filesystem paths which may be used to
        resolve relative mojom file paths.
    output_root_path: An absolute filesystem path which will service as the root
        for all emitted artifacts. Artifacts produced from a given mojom file
        are based on the mojom's relative path, rebased onto this path.
        Additionally, the script expects this root to contain already-generated
        modules for any transitive dependencies not listed in mojom_files.
    module_root_paths: A list of absolute filesystem paths which contain
        already-generated modules for any non-transitive dependencies.
    enabled_features: A list of enabled feature names, controlling which AST
        nodes are filtered by [EnableIf] or [EnableIfNot] attributes.
    module_metadata: A list of 2-tuples representing metadata key-value pairs to
        attach to each compiled module output.

  Returns:
    None.

    Upon completion, a mojom-module file will be saved for each input mojom.
  """
  assert input_root_paths
  assert output_root_path

  loaded_mojom_asts = {}
  loaded_modules = {}
  input_dependencies = defaultdict(set)
  mojom_files_to_parse = dict((os.path.normcase(abs_path),
                               RebaseAbsolutePath(abs_path, input_root_paths))
                              for abs_path in mojom_files)
  abs_paths = dict(
      (path, abs_path) for abs_path, path in mojom_files_to_parse.items())

  logging.info('Parsing %d .mojom into ASTs', len(mojom_files_to_parse))
  map_args = ((mojom_abspath, enabled_features)
              for mojom_abspath in mojom_files_to_parse)
  for mojom_abspath, ast in _Shard(_ParseAstHelper, map_args):
    loaded_mojom_asts[mojom_abspath] = ast

  logging.info('Processing dependencies')
  for mojom_abspath, ast in sorted(loaded_mojom_asts.items()):
    invalid_imports = []
    for imp in ast.import_list:
      import_abspath = _ResolveRelativeImportPath(imp.import_filename,
                                                  input_root_paths)
      if allowed_imports and import_abspath not in allowed_imports:
        invalid_imports.append(imp.import_filename)

      abs_paths[imp.import_filename] = import_abspath
      if import_abspath in mojom_files_to_parse:
        # This import is in the input list, so we're going to translate it
        # into a module below; however it's also a dependency of another input
        # module. We retain record of dependencies to help with input
        # processing later.
        input_dependencies[mojom_abspath].add(
            (import_abspath, imp.import_filename))
      elif import_abspath not in loaded_modules:
        # We have an import that isn't being parsed right now. It must already
        # be parsed and have a module file sitting in a corresponding output
        # location.
        module_path = _GetModuleFilename(imp.import_filename)
        module_abspath = _ResolveRelativeImportPath(
            module_path, module_root_paths + [output_root_path])
        with open(module_abspath, 'rb') as module_file:
          loaded_modules[import_abspath] = module.Module.Load(module_file)

    if invalid_imports:
      raise ValueError(
          '\nThe file %s imports the following files not allowed by build '
          'dependencies:\n\n%s\n' % (mojom_abspath, '\n'.join(invalid_imports)))
  logging.info('Loaded %d modules from dependencies', len(loaded_modules))

  # At this point all transitive imports not listed as inputs have been loaded
  # and we have a complete dependency tree of the unprocessed inputs. Now we can
  # load all the inputs, resolving dependencies among them recursively as we go.
  logging.info('Ensuring inputs are loaded')
  num_existing_modules_loaded = len(loaded_modules)
  for mojom_abspath, mojom_path in mojom_files_to_parse.items():
    _EnsureInputLoaded(mojom_abspath, mojom_path, abs_paths, loaded_mojom_asts,
                       input_dependencies, loaded_modules, module_metadata)
  assert (num_existing_modules_loaded +
          len(mojom_files_to_parse) == len(loaded_modules))

  # Now we have fully translated modules for every input and every transitive
  # dependency. We can dump the modules to disk for other tools to use.
  logging.info('Serializing %d modules', len(mojom_files_to_parse))

  # Windows does not use fork() for multiprocessing, so we'd need to pass
  # loaded_module via IPC rather than via globals. Doing so is slower than not
  # using multiprocessing.
  _SerializeHelper.loaded_modules = loaded_modules
  _SerializeHelper.output_root_path = output_root_path
  # Doesn't seem to help past 4. Perhaps IO bound here?
  processes = 4 if _MULTIPROCESSING_USES_FORK else 0
  map_args = mojom_files_to_parse.items()
  for _ in _Shard(_SerializeHelper, map_args, processes=processes):
    pass


def Run(command_line):
  debug_logging = os.environ.get('MOJOM_PARSER_DEBUG', '0') != '0'
  logging.basicConfig(level=logging.DEBUG if debug_logging else logging.WARNING,
                      format='%(levelname).1s %(relativeCreated)6d %(message)s')
  logging.info('Started (%s)', os.path.basename(sys.argv[0]))

  arg_parser = argparse.ArgumentParser(
      description="""
Parses one or more mojom files and produces corresponding module outputs fully
describing the definitions therein. The output is exhaustive, stable, and
sufficient for another tool to consume and emit e.g. usable language
bindings based on the original mojoms.""",
      epilog="""
Note that each transitive import dependency reachable from the input mojoms must
either also be listed as an input or must have its corresponding compiled module
already present in the provided output root.""")

  arg_parser.add_argument(
      '--input-root',
      default=[],
      action='append',
      metavar='ROOT',
      dest='input_root_paths',
      help='Adds ROOT to the set of root paths against which relative input '
      'paths should be resolved. Provided root paths are always searched '
      'in order from longest absolute path to shortest.')
  arg_parser.add_argument(
      '--output-root',
      action='store',
      required=True,
      dest='output_root_path',
      metavar='ROOT',
      help='Use ROOT as the root path in which the parser should emit compiled '
      'modules for each processed input mojom. The path of emitted module is '
      'based on the relative input path, rebased onto this root. Note that '
      'ROOT is also searched for existing modules of any transitive imports '
      'which were not included in the set of inputs.')
  arg_parser.add_argument(
      '--module-root',
      default=[],
      action='append',
      metavar='ROOT',
      dest='module_root_paths',
      help='Adds ROOT to the set of root paths to search for existing modules '
      'of non-transitive imports. Provided root paths are always searched in '
      'order from longest absolute path to shortest.')
  arg_parser.add_argument(
      '--mojoms',
      nargs='+',
      dest='mojom_files',
      default=[],
      metavar='MOJOM_FILE',
      help='Input mojom filename(s). Each filename must be either an absolute '
      'path which falls within one of the given input or output roots, or a '
      'relative path the parser will attempt to resolve using each of those '
      'roots in unspecified order.')
  arg_parser.add_argument(
      '--mojom-file-list',
      action='store',
      metavar='LIST_FILENAME',
      help='Input file whose contents are a list of mojoms to process. This '
      'may be provided in lieu of --mojoms to avoid hitting command line '
      'length limtations')
  arg_parser.add_argument(
      '--enable-feature',
      dest='enabled_features',
      default=[],
      action='append',
      metavar='FEATURE',
      help='Enables a named feature when parsing the given mojoms. Features '
      'are identified by arbitrary string values. Specifying this flag with a '
      'given FEATURE name will cause the parser to process any syntax elements '
      'tagged with an [EnableIf=FEATURE] or [EnableIfNot] attribute. If this '
      'flag is not provided for a given FEATURE, such tagged elements are '
      'discarded by the parser and will not be present in the compiled output.')
  arg_parser.add_argument(
      '--check-imports',
      dest='build_metadata_filename',
      action='store',
      metavar='METADATA_FILENAME',
      help='Instructs the parser to check imports against a set of allowed '
      'imports. Allowed imports are based on build metadata within '
      'METADATA_FILENAME. This is a JSON file with a `sources` key listing '
      'paths to the set of input mojom files being processed by this parser '
      'run, and a `deps` key listing paths to metadata files for any '
      'dependencies of these inputs. This feature can be used to implement '
      'build-time dependency checking for mojom imports, where each build '
      'metadata file corresponds to a build target in the dependency graph of '
      'a typical build system.')
  arg_parser.add_argument(
      '--add-module-metadata',
      dest='module_metadata',
      default=[],
      action='append',
      metavar='KEY=VALUE',
      help='Adds a metadata key-value pair to the output module. This can be '
      'used by build toolchains to augment parsed mojom modules with product-'
      'specific metadata for later extraction and use by custom bindings '
      'generators.')

  args, _ = arg_parser.parse_known_args(command_line)
  if args.mojom_file_list:
    with open(args.mojom_file_list) as f:
      args.mojom_files.extend(f.read().split())

  if not args.mojom_files:
    raise ValueError(
        'Must list at least one mojom file via --mojoms or --mojom-file-list')

  mojom_files = list(map(os.path.abspath, args.mojom_files))
  input_roots = list(map(os.path.abspath, args.input_root_paths))
  output_root = os.path.abspath(args.output_root_path)
  module_roots = list(map(os.path.abspath, args.module_root_paths))

  if args.build_metadata_filename:
    allowed_imports = _CollectAllowedImportsFromBuildMetadata(
        args.build_metadata_filename)
  else:
    allowed_imports = None

  module_metadata = list(
      map(lambda kvp: tuple(kvp.split('=')), args.module_metadata))
  _ParseMojoms(mojom_files, input_roots, output_root, module_roots,
               args.enabled_features, module_metadata, allowed_imports)
  logging.info('Finished')


if __name__ == '__main__':
  Run(sys.argv[1:])
  # Exit without running GC, which can save multiple seconds due to the large
  # number of object created. But flush is necessary as os._exit doesn't do
  # that.
  sys.stdout.flush()
  sys.stderr.flush()
  os._exit(0)



# SPDX-License-Identifier: GPL-2.0-or-later
# Copyright (C) 2022, Tomi Valkeinen <tomi.valkeinen@ideasonboard.com>

import pykms
import selectors
import sys


class KMSRenderer:
    def __init__(self, state):
        self.state = state

        self.cm = state.cm
        self.contexts = state.contexts
        self.running = False

        card = pykms.Card()

        res = pykms.ResourceManager(card)
        conn = res.reserve_connector()
        crtc = res.reserve_crtc(conn)
        mode = conn.get_default_mode()
        modeb = mode.to_blob(card)

        req = pykms.AtomicReq(card)
        req.add_connector(conn, crtc)
        req.add_crtc(crtc, modeb)
        r = req.commit_sync(allow_modeset=True)
        assert(r == 0)

        self.card = card
        self.resman = res
        self.crtc = crtc
        self.mode = mode

        self.bufqueue = []
        self.current = None
        self.next = None
        self.cam_2_drm = {}

    # KMS

    def close(self):
        req = pykms.AtomicReq(self.card)
        for s in self.streams:
            req.add_plane(s['plane'], None, None, dst=(0, 0, 0, 0))
        req.commit()

    def add_plane(self, req, stream, fb):
        s = next(s for s in self.streams if s['stream'] == stream)
        idx = s['idx']
        plane = s['plane']

        if idx % 2 == 0:
            x = 0
        else:
            x = self.mode.hdisplay - fb.width

        if idx // 2 == 0:
            y = 0
        else:
            y = self.mode.vdisplay - fb.height

        req.add_plane(plane, fb, self.crtc, dst=(x, y, fb.width, fb.height))

    def apply_request(self, drmreq):

        buffers = drmreq['camreq'].buffers

        req = pykms.AtomicReq(self.card)

        for stream, fb in buffers.items():
            drmfb = self.cam_2_drm.get(fb, None)
            self.add_plane(req, stream, drmfb)

        req.commit()

    def handle_page_flip(self, frame, time):
        old = self.current
        self.current = self.next

        if len(self.bufqueue) > 0:
            self.next = self.bufqueue.pop(0)
        else:
            self.next = None

        if self.next:
            drmreq = self.next

            self.apply_request(drmreq)

        if old:
            req = old['camreq']
            ctx = old['camctx']
            self.state.request_processed(ctx, req)

    def queue(self, drmreq):
        if not self.next:
            self.next = drmreq
            self.apply_request(drmreq)
        else:
            self.bufqueue.append(drmreq)

    # libcamera

    def setup(self):
        self.streams = []

        idx = 0
        for ctx in self.contexts:
            for stream in ctx.streams:

                cfg = stream.configuration
                fmt = cfg.pixel_format
                fmt = pykms.PixelFormat(fmt.fourcc)

                plane = self.resman.reserve_generic_plane(self.crtc, fmt)
                assert(plane is not None)

                self.streams.append({
                    'idx': idx,
                    'stream': stream,
                    'plane': plane,
                    'fmt': fmt,
                    'size': cfg.size,
                })

                for fb in ctx.allocator.buffers(stream):
                    w = cfg.size.width
                    h = cfg.size.height
                    fds = []
                    strides = []
                    offsets = []
                    for plane in fb.planes:
                        fds.append(plane.fd)
                        strides.append(cfg.stride)
                        offsets.append(plane.offset)

                    drmfb = pykms.DmabufFramebuffer(self.card, w, h, fmt,
                                                    fds, strides, offsets)
                    self.cam_2_drm[fb] = drmfb

                idx += 1

    def readdrm(self, fileobj):
        for ev in self.card.read_events():
            if ev.type == pykms.DrmEventType.FLIP_COMPLETE:
                self.handle_page_flip(ev.seq, ev.time)

    def readcam(self, fd):
        self.running = self.state.event_handler()

    def readkey(self, fileobj):
        sys.stdin.readline()
        self.running = False

    def run(self):
        print('Capturing...')

        self.running = True

        sel = selectors.DefaultSelector()
        sel.register(self.card.fd, selectors.EVENT_READ, self.readdrm)
        sel.register(self.cm.event_fd, selectors.EVENT_READ, self.readcam)
        sel.register(sys.stdin, selectors.EVENT_READ, self.readkey)

        print('Press enter to exit')

        while self.running:
            events = sel.select()
            for key, mask in events:
                callback = key.data
                callback(key.fileobj)

        print('Exiting...')

    def request_handler(self, ctx, req):

        drmreq = {
            'camctx': ctx,
            'camreq': req,
        }

        self.queue(drmreq)



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0-or-later
#
# Generate Python bindings controls from YAML

import argparse
import string
import sys
import yaml


def find_common_prefix(strings):
    prefix = strings[0]

    for string in strings[1:]:
        while string[:len(prefix)] != prefix and prefix:
            prefix = prefix[:len(prefix) - 1]
        if not prefix:
            break

    return prefix


def generate_py(controls, mode):
    out = ''

    vendors_class_def = []
    vendor_defs = []
    vendors = []
    for vendor, ctrl_list in controls.items():
        for ctrls in ctrl_list:
            name, ctrl = ctrls.popitem()

            if vendor not in vendors and vendor != 'libcamera':
                vendor_mode_str = f'{vendor.capitalize()}{mode.capitalize()}'
                vendors_class_def.append('class Py{}\n{{\n}};\n'.format(vendor_mode_str))
                vendor_defs.append('\tauto {} = py::class_<Py{}>(controls, \"{}\");'.format(vendor, vendor_mode_str, vendor))
                vendors.append(vendor)

            if vendor != 'libcamera':
                ns = 'libcamera::{}::{}::'.format(mode, vendor)
                container = vendor
            else:
                ns = 'libcamera::{}::'.format(mode)
                container = 'controls'

            out += f'\t{container}.def_readonly_static("{name}", static_cast<const libcamera::ControlId *>(&{ns}{name}));\n\n'

            enum = ctrl.get('enum')
            if not enum:
                continue

            cpp_enum = name + 'Enum'

            out += '\tpy::enum_<{}{}>({}, \"{}\")\n'.format(ns, cpp_enum, container, cpp_enum)

            if mode == 'controls':
                # Adjustments for controls
                if name == 'LensShadingMapMode':
                    prefix = 'LensShadingMapMode'
                else:
                    prefix = find_common_prefix([e['name'] for e in enum])
            else:
                # Adjustments for properties
                prefix = find_common_prefix([e['name'] for e in enum])

            for entry in enum:
                cpp_enum = entry['name']
                py_enum = entry['name'][len(prefix):]

                out += '\t\t.value(\"{}\", {}{})\n'.format(py_enum, ns, cpp_enum)

            out += '\t;\n\n'

    return {'controls': out,
            'vendors_class_def': '\n'.join(vendors_class_def),
            'vendors_defs': '\n'.join(vendor_defs)}


def fill_template(template, data):
    template = open(template, 'rb').read()
    template = template.decode('utf-8')
    template = string.Template(template)
    return template.substitute(data)


def main(argv):
    # Parse command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', '-m', type=str, required=True,
                        help='Mode is either "controls" or "properties"')
    parser.add_argument('--output', '-o', metavar='file', type=str,
                        help='Output file name. Defaults to standard output if not specified.')
    parser.add_argument('--template', '-t', type=str, required=True,
                        help='Template file name.')
    parser.add_argument('input', type=str, nargs='+',
                        help='Input file name.')
    args = parser.parse_args(argv[1:])

    if args.mode not in ['controls', 'properties']:
        print(f'Invalid mode option "{args.mode}"', file=sys.stderr)
        return -1

    controls = {}
    for input in args.input:
        data = open(input, 'rb').read()
        vendor = yaml.safe_load(data)['vendor']
        controls[vendor] = yaml.safe_load(data)['controls']

    data = generate_py(controls, args.mode)

    data = fill_template(args.template, data)

    if args.output:
        output = open(args.output, 'wb')
        output.write(data.encode('utf-8'))
        output.close()
    else:
        sys.stdout.write(data)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))



# Copyright 2015 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import errno
import os.path
import sys


def _GetDirAbove(dirname):
  """Returns the directory "above" this file containing |dirname| (which must
  also be "above" this file)."""
  path = os.path.abspath(__file__)
  while True:
    path, tail = os.path.split(path)
    if not tail:
      return None
    if tail == dirname:
      return path


def EnsureDirectoryExists(path, always_try_to_create=False):
  """A wrapper for os.makedirs that does not error if the directory already
  exists. A different process could be racing to create this directory."""

  if not os.path.exists(path) or always_try_to_create:
    try:
      os.makedirs(path)
    except OSError as e:
      # There may have been a race to create this directory.
      if e.errno != errno.EEXIST:
        raise


def AddLocalRepoThirdPartyDirToModulePath():
  """Helper function to find the top-level directory of this script's repository
  assuming the script falls somewhere within a 'mojo' directory, and insert the
  top-level 'third_party' directory early in the module search path. Used to
  ensure that third-party dependencies provided within the repository itself
  (e.g. Chromium sources include snapshots of jinja2 and ply) are preferred over
  locally installed system library packages."""
  toplevel_dir = _GetDirAbove('mojo')
  if toplevel_dir:
    sys.path.insert(1, os.path.join(toplevel_dir, 'third_party'))



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# ctt_ransac.py - camera tuning tool RANSAC selector for Macbeth chart locator

import numpy as np

scale = 2


"""
constructs normalised macbeth chart corners for ransac algorithm
"""
def get_square_verts(c_err=0.05, scale=scale):
    """
    define macbeth chart corners
    """
    b_bord_x, b_bord_y = scale*8.5, scale*13
    s_bord = 6*scale
    side = 41*scale
    x_max = side*6 + 5*s_bord + 2*b_bord_x
    y_max = side*4 + 3*s_bord + 2*b_bord_y
    c1 = (0, 0)
    c2 = (0, y_max)
    c3 = (x_max, y_max)
    c4 = (x_max, 0)
    mac_norm = np.array((c1, c2, c3, c4), np.float32)
    mac_norm = np.array([mac_norm])

    square_verts = []
    square_0 = np.array(((0, 0), (0, side),
                         (side, side), (side, 0)), np.float32)
    offset_0 = np.array((b_bord_x, b_bord_y), np.float32)
    c_off = side * c_err
    offset_cont = np.array(((c_off, c_off), (c_off, -c_off),
                            (-c_off, -c_off), (-c_off, c_off)), np.float32)
    square_0 += offset_0
    square_0 += offset_cont
    """
    define macbeth square corners
    """
    for i in range(6):
        shift_i = np.array(((i*side, 0), (i*side, 0),
                            (i*side, 0), (i*side, 0)), np.float32)
        shift_bord = np.array(((i*s_bord, 0), (i*s_bord, 0),
                               (i*s_bord, 0), (i*s_bord, 0)), np.float32)
        square_i = square_0 + shift_i + shift_bord
        for j in range(4):
            shift_j = np.array(((0, j*side), (0, j*side),
                                (0, j*side), (0, j*side)), np.float32)
            shift_bord = np.array(((0, j*s_bord),
                                   (0, j*s_bord), (0, j*s_bord),
                                   (0, j*s_bord)), np.float32)
            square_j = square_i + shift_j + shift_bord
            square_verts.append(square_j)
    # print('square_verts')
    # print(square_verts)
    return np.array(square_verts, np.float32), mac_norm


def get_square_centres(c_err=0.05, scale=scale):
    """
    define macbeth square centres
    """
    verts, mac_norm = get_square_verts(c_err, scale=scale)

    centres = np.mean(verts, axis=1)
    # print('centres')
    # print(centres)
    return np.array(centres, np.float32)



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
# Copyright (C) 2022, Paul Elder <paul.elder@ideasonboard.com>
#
# raspberrypi.py - ALSC module for tuning Raspberry Pi

from .lsc import LSC

import libtuning as lt
import libtuning.utils as utils

from numbers import Number
import numpy as np


class ALSCRaspberryPi(LSC):
    # Override the type name so that the parser can match the entry in the
    # config file.
    type = 'alsc'
    hr_name = 'ALSC (Raspberry Pi)'
    out_name = 'rpi.alsc'
    compatible = ['raspberrypi']

    def __init__(self, *,
                 do_color: lt.Param,
                 luminance_strength: lt.Param,
                 **kwargs):
        super().__init__(**kwargs)

        self.do_color = do_color
        self.luminance_strength = luminance_strength

        self.output_range = (0, 3.999)

    def validate_config(self, config: dict) -> bool:
        if self not in config:
            utils.eprint(f'{self.type} not in config')
            return False

        valid = True

        conf = config[self]

        lum_key = self.luminance_strength.name
        color_key = self.do_color.name

        if lum_key not in conf and self.luminance_strength.required:
            utils.eprint(f'{lum_key} is not in config')
            valid = False

        if lum_key in conf and (conf[lum_key] < 0 or conf[lum_key] > 1):
            utils.eprint(f'Warning: {lum_key} is not in range [0, 1]; defaulting to 0.5')

        if color_key not in conf and self.do_color.required:
            utils.eprint(f'{color_key} is not in config')
            valid = False

        return valid

    # @return Image color temperature, flattened array of red calibration table
    #         (containing {sector size} elements), flattened array of blue
    #         calibration table, flattened array of green calibration
    #         table

    def _do_single_alsc(self, image: lt.Image, do_alsc_colour: bool):
        average_green = np.mean((image.channels[lt.Color.GR:lt.Color.GB + 1]), axis=0)

        cg, g = self._lsc_single_channel(average_green, image)

        if not do_alsc_colour:
            return image.color, None, None, cg.flatten()

        cr, _ = self._lsc_single_channel(image.channels[lt.Color.R], image, g)
        cb, _ = self._lsc_single_channel(image.channels[lt.Color.B], image, g)

        # \todo implement debug

        return image.color, cr.flatten(), cb.flatten(), cg.flatten()

    # @return Red shading table, Blue shading table, Green shading table,
    #         number of images processed

    def _do_all_alsc(self, images: list, do_alsc_colour: bool, general_conf: dict) -> (list, list, list, Number, int):
        # List of colour temperatures
        list_col = []
        # Associated calibration tables
        list_cr = []
        list_cb = []
        list_cg = []
        count = 0
        for image in self._enumerate_lsc_images(images):
            col, cr, cb, cg = self._do_single_alsc(image, do_alsc_colour)
            list_col.append(col)
            list_cr.append(cr)
            list_cb.append(cb)
            list_cg.append(cg)
            count += 1

        # Convert to numpy array for data manipulation
        list_col = np.array(list_col)
        list_cr = np.array(list_cr)
        list_cb = np.array(list_cb)
        list_cg = np.array(list_cg)

        cal_cr_list = []
        cal_cb_list = []

        # Note: Calculation of average corners and center of the shading tables
        # has been removed (which ctt had, as it was unused)

        # Average all values for luminance shading and return one table for all temperatures
        lum_lut = list(np.round(np.mean(list_cg, axis=0), 3))

        if not do_alsc_colour:
            return None, None, lum_lut, count

        for ct in sorted(set(list_col)):
            # Average tables for the same colour temperature
            indices = np.where(list_col == ct)
            ct = int(ct)
            t_r = np.round(np.mean(list_cr[indices], axis=0), 3)
            t_b = np.round(np.mean(list_cb[indices], axis=0), 3)

            cr_dict = {
                'ct': ct,
                'table': list(t_r)
            }
            cb_dict = {
                'ct': ct,
                'table': list(t_b)
            }
            cal_cr_list.append(cr_dict)
            cal_cb_list.append(cb_dict)

        return cal_cr_list, cal_cb_list, lum_lut, count

    # @brief Calculate sigma from two adjacent gain tables
    def _calcSigma(self, g1, g2):
        g1 = np.reshape(g1, self.sector_shape[::-1])
        g2 = np.reshape(g2, self.sector_shape[::-1])

        # Apply gains to gain table
        gg = g1 / g2
        if np.mean(gg) < 1:
            gg = 1 / gg

        # For each internal patch, compute average difference between it and
        # its 4 neighbours, then append to list
        diffs = []
        for i in range(self.sector_shape[1] - 2):
            for j in range(self.sector_shape[0] - 2):
                # Indexing is incremented by 1 since all patches on borders are
                # not counted
                diff = np.abs(gg[i + 1][j + 1] - gg[i][j + 1])
                diff += np.abs(gg[i + 1][j + 1] - gg[i + 2][j + 1])
                diff += np.abs(gg[i + 1][j + 1] - gg[i + 1][j])
                diff += np.abs(gg[i + 1][j + 1] - gg[i + 1][j + 2])
                diffs.append(diff / 4)

        mean_diff = np.mean(diffs)
        return np.round(mean_diff, 5)

    # @brief Obtains sigmas for red and blue, effectively a measure of the
    # 'error'
    def _get_sigma(self, cal_cr_list, cal_cb_list):
        # Provided colour alsc tables were generated for two different colour
        # temperatures sigma is calculated by comparing two calibration temperatures
        # adjacent in colour space

        color_temps = [cal['ct'] for cal in cal_cr_list]

        # Calculate sigmas for each adjacent color_temps and return worst one
        sigma_rs = []
        sigma_bs = []
        for i in range(len(color_temps) - 1):
            sigma_rs.append(self._calcSigma(cal_cr_list[i]['table'], cal_cr_list[i + 1]['table']))
            sigma_bs.append(self._calcSigma(cal_cb_list[i]['table'], cal_cb_list[i + 1]['table']))

        # Return maximum sigmas, not necessarily from the same colour
        # temperature interval
        sigma_r = max(sigma_rs) if sigma_rs else 0.005
        sigma_b = max(sigma_bs) if sigma_bs else 0.005

        return sigma_r, sigma_b

    def process(self, config: dict, images: list, outputs: dict) -> dict:
        output = {
            'omega': 1.3,
            'n_iter': 100,
            'luminance_strength': 0.7
        }

        conf = config[self]
        general_conf = config['general']

        do_alsc_colour = self.do_color.get_value(conf)

        # \todo I have no idea where this input parameter is used
        luminance_strength = self.luminance_strength.get_value(conf)
        if luminance_strength < 0 or luminance_strength > 1:
            luminance_strength = 0.5

        output['luminance_strength'] = luminance_strength

        # \todo Validate images from greyscale camera and force grescale mode
        # \todo Debug functionality

        alsc_out = self._do_all_alsc(images, do_alsc_colour, general_conf)
        # \todo Handle the second green lut
        cal_cr_list, cal_cb_list, luminance_lut, count = alsc_out

        if not do_alsc_colour:
            output['luminance_lut'] = luminance_lut
            output['n_iter'] = 0
            return output

        output['calibrations_Cr'] = cal_cr_list
        output['calibrations_Cb'] = cal_cb_list
        output['luminance_lut'] = luminance_lut

        # The sigmas determine the strength of the adaptive algorithm, that
        # cleans up any lens shading that has slipped through the alsc. These
        # are determined by measuring a 'worst-case' difference between two
        # alsc tables that are adjacent in colour space. If, however, only one
        # colour temperature has been provided, then this difference can not be
        # computed as only one table is available.
        # To determine the sigmas you would have to estimate the error of an
        # alsc table with only the image it was taken on as a check. To avoid
        # circularity, dfault exaggerated sigmas are used, which can result in
        # too much alsc and is therefore not advised.
        # In general, just take another alsc picture at another colour
        # temperature!

        if count == 1:
            output['sigma'] = 0.005
            output['sigma_Cb'] = 0.005
            utils.eprint('Warning: Only one alsc calibration found; standard sigmas used for adaptive algorithm.')
            return output

        # Obtain worst-case scenario residual sigmas
        sigma_r, sigma_b = self._get_sigma(cal_cr_list, cal_cb_list)
        output['sigma'] = np.round(sigma_r, 5)
        output['sigma_Cb'] = np.round(sigma_b, 5)

        return output



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright 2022 Raspberry Pi Ltd
#
# raspberrypi_output.py - Generate tuning file in Raspberry Pi's json format
#
# (Copied from ctt_pretty_print_json.py)

from .generator import Generator

import json
from pathlib import Path
import textwrap


class Encoder(json.JSONEncoder):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.indentation_level = 0
        self.hard_break = 120
        self.custom_elems = {
            'table': 16,
            'luminance_lut': 16,
            'ct_curve': 3,
            'ccm': 3,
            'gamma_curve': 2,
            'y_target': 2,
            'prior': 2
        }

    def encode(self, o, node_key=None):
        if isinstance(o, (list, tuple)):
            # Check if we are a flat list of numbers.
            if not any(isinstance(el, (list, tuple, dict)) for el in o):
                s = ', '.join(json.dumps(el) for el in o)
                if node_key in self.custom_elems.keys():
                    # Special case handling to specify number of elements in a row for tables, ccm, etc.
                    self.indentation_level += 1
                    sl = s.split(', ')
                    num = self.custom_elems[node_key]
                    chunk = [self.indent_str + ', '.join(sl[x:x + num]) for x in range(0, len(sl), num)]
                    t = ',\n'.join(chunk)
                    self.indentation_level -= 1
                    output = f'\n{self.indent_str}[\n{t}\n{self.indent_str}]'
                elif len(s) > self.hard_break - len(self.indent_str):
                    # Break a long list with wraps.
                    self.indentation_level += 1
                    t = textwrap.fill(s, self.hard_break, break_long_words=False,
                                      initial_indent=self.indent_str, subsequent_indent=self.indent_str)
                    self.indentation_level -= 1
                    output = f'\n{self.indent_str}[\n{t}\n{self.indent_str}]'
                else:
                    # Smaller lists can remain on a single line.
                    output = f' [ {s} ]'
                return output
            else:
                # Sub-structures in the list case.
                self.indentation_level += 1
                output = [self.indent_str + self.encode(el) for el in o]
                self.indentation_level -= 1
                output = ',\n'.join(output)
                return f' [\n{output}\n{self.indent_str}]'

        elif isinstance(o, dict):
            self.indentation_level += 1
            output = []
            for k, v in o.items():
                if isinstance(v, dict) and len(v) == 0:
                    # Empty config block special case.
                    output.append(self.indent_str + f'{json.dumps(k)}: {{ }}')
                else:
                    # Only linebreak if the next node is a config block.
                    sep = f'\n{self.indent_str}' if isinstance(v, dict) else ''
                    output.append(self.indent_str + f'{json.dumps(k)}:{sep}{self.encode(v, k)}')
            output = ',\n'.join(output)
            self.indentation_level -= 1
            return f'{{\n{output}\n{self.indent_str}}}'

        else:
            return ' ' + json.dumps(o)

    @property
    def indent_str(self) -> str:
        return ' ' * self.indentation_level * self.indent

    def iterencode(self, o, **kwargs):
        return self.encode(o)


class RaspberryPiOutput(Generator):
    def __init__(self):
        super().__init__()

    def _pretty_print(self, in_json: dict) -> str:

        if 'version' not in in_json or \
           'target' not in in_json or \
           'algorithms' not in in_json or \
           in_json['version'] < 2.0:
            raise RuntimeError('Incompatible JSON dictionary has been provided')

        return json.dumps(in_json, cls=Encoder, indent=4, sort_keys=False)

    def write(self, output_file: Path, output_dict: dict, output_order: list):
        # Write json dictionary to file using ctt's version 2 format
        out_json = {
            "version": 2.0,
            'target': 'bcm2835',
            "algorithms": [{f'{module.out_name}': output_dict[module]} for module in output_order]
        }

        with open(output_file, 'w') as f:
            f.write(self._pretty_print(out_json))



# Copyright 2023 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Validate mojo runtime feature guarded interfaces are nullable."""

import mojom.generate.check as check
import mojom.generate.module as module


class Check(check.Check):
  def __init__(self, *args, **kwargs):
    super(Check, self).__init__(*args, **kwargs)

  # `param` is an Interface of some sort.
  def _CheckNonNullableFeatureGuardedInterface(self, kind):
    # Only need to validate interface if it has a RuntimeFeature
    if not kind.kind.runtime_feature:
      return
    # Nullable (optional) is ok as the interface expects they might not be sent.
    if kind.is_nullable:
      return
    interface = kind.kind.mojom_name
    raise check.CheckException(
        self.module,
        f"interface {interface} has a RuntimeFeature but is not nullable")

  # `param` can be a lot of things so check if it is a remote/receiver.
  # Array/Map must be recursed into.
  def _CheckFieldOrParam(self, kind):
    if module.IsAnyInterfaceKind(kind):
      self._CheckNonNullableFeatureGuardedInterface(kind)
    if module.IsArrayKind(kind):
      self._CheckFieldOrParam(kind.kind)
    if module.IsMapKind(kind):
      self._CheckFieldOrParam(kind.key_kind)
      self._CheckFieldOrParam(kind.value_kind)

  def _CheckInterfaceFeatures(self, interface):
    for method in interface.methods:
      for param in method.parameters:
        self._CheckFieldOrParam(param.kind)
      if method.response_parameters:
        for param in method.response_parameters:
          self._CheckFieldOrParam(param.kind)

  def _CheckStructFeatures(self, struct):
    for field in struct.fields:
      self._CheckFieldOrParam(field.kind)

  def _CheckUnionFeatures(self, union):
    for field in union.fields:
      self._CheckFieldOrParam(field.kind)

  def CheckModule(self):
    """Validate that any runtime feature guarded interfaces that might be passed
    over mojo are nullable."""
    for interface in self.module.interfaces:
      self._CheckInterfaceFeatures(interface)
    for struct in self.module.structs:
      self._CheckStructFeatures(struct)
    for union in self.module.unions:
      self._CheckUnionFeatures(union)



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# ctt_noise.py - camera tuning tool noise calibration

from ctt_image_load import *
import matplotlib.pyplot as plt


"""
Find noise standard deviation and fit to model:

    noise std = a + b*sqrt(pixel mean)
"""
def noise(Cam, Img, plot):
    Cam.log += '\nProcessing image: {}'.format(Img.name)
    stds = []
    means = []
    """
    iterate through macbeth square patches
    """
    for ch_patches in Img.patches:
        for patch in ch_patches:
            """
            renormalise patch
            """
            patch = np.array(patch)
            patch = (patch-Img.blacklevel_16)/Img.againQ8_norm
            std = np.std(patch)
            mean = np.mean(patch)
            stds.append(std)
            means.append(mean)

    """
    clean data and ensure all means are above 0
    """
    stds = np.array(stds)
    means = np.array(means)
    means = np.clip(np.array(means), 0, None)
    sq_means = np.sqrt(means)

    """
    least squares fit model
    """
    fit = np.polyfit(sq_means, stds, 1)
    Cam.log += '\nBlack level = {}'.format(Img.blacklevel_16)
    Cam.log += '\nNoise profile: offset = {}'.format(int(fit[1]))
    Cam.log += ' slope = {:.3f}'.format(fit[0])
    """
    remove any values further than std from the fit

    anomalies most likely caused by:
    > ucharacteristically noisy white patch
    > saturation in the white patch
    """
    fit_score = np.abs(stds - fit[0]*sq_means - fit[1])
    fit_std = np.std(stds)
    fit_score_norm = fit_score - fit_std
    anom_ind = np.where(fit_score_norm > 1)
    fit_score_norm.sort()
    sq_means_clean = np.delete(sq_means, anom_ind)
    stds_clean = np.delete(stds, anom_ind)
    removed = len(stds) - len(stds_clean)
    if removed != 0:
        Cam.log += '\nIdentified and removed {} anomalies.'.format(removed)
        Cam.log += '\nRecalculating fit'
        """
        recalculate fit with outliers removed
        """
        fit = np.polyfit(sq_means_clean, stds_clean, 1)
        Cam.log += '\nNoise profile: offset = {}'.format(int(fit[1]))
        Cam.log += ' slope = {:.3f}'.format(fit[0])

    """
    if fit const is < 0 then force through 0 by
    dividing by sq_means and fitting poly order 0
    """
    corrected = 0
    if fit[1] < 0:
        corrected = 1
        ones = np.ones(len(means))
        y_data = stds/sq_means
        fit2 = np.polyfit(ones, y_data, 0)
        Cam.log += '\nOffset below zero. Fit recalculated with zero offset'
        Cam.log += '\nNoise profile: offset = 0'
        Cam.log += ' slope = {:.3f}'.format(fit2[0])
        # print('new fit')
        # print(fit2)

    """
    plot fit for debug
    """
    if plot:
        x = np.arange(sq_means.max()//0.88)
        fit_plot = x*fit[0] + fit[1]
        plt.scatter(sq_means, stds, label='data', color='blue')
        plt.scatter(sq_means[anom_ind], stds[anom_ind], color='orange', label='anomalies')
        plt.plot(x, fit_plot, label='fit', color='red', ls=':')
        if fit[1] < 0:
            fit_plot_2 = x*fit2[0]
            plt.plot(x, fit_plot_2, label='fit 0 intercept', color='green', ls='--')
        plt.plot(0, 0)
        plt.title('Noise Plot\nImg: {}'.format(Img.str))
        plt.legend(loc='upper left')
        plt.xlabel('Sqrt Pixel Value')
        plt.ylabel('Noise Standard Deviation')
        plt.grid()
        plt.show()
    """
    End of plotting code
    """

    """
    format output to include forced 0 constant
    """
    Cam.log += '\n'
    if corrected:
        fit = [fit2[0], 0]
        return fit

    else:
        return fit



# SPDX-License-Identifier: BSD-2-Clause
#
# Copyright (C) 2019, Raspberry Pi Ltd
#
# ctt_ccm.py - camera tuning tool for CCM (colour correction matrix)

from ctt_image_load import *
from ctt_awb import get_alsc_patches
import colors
from scipy.optimize import minimize
from ctt_visualise import visualise_macbeth_chart
import numpy as np
"""
takes 8-bit macbeth chart values, degammas and returns 16 bit
"""

'''
This program has many options from which to derive the color matrix from.
The first is average. This minimises the average delta E across all patches of
the macbeth chart. Testing across all cameras yeilded this as the most color
accurate and vivid. Other options are avalible however.
Maximum minimises the maximum Delta E of the patches. It iterates through till
a minimum maximum is found (so that there is
not one patch that deviates wildly.)
This yields generally good results but overall the colors are less accurate
Have a fiddle with maximum and see what you think.
The final option allows you to select the patches for which to average across.
This means that you can bias certain patches, for instance if you want the
reds to be more accurate.
'''

matrix_selection_types = ["average", "maximum", "patches"]
typenum = 0  # select from array above, 0 = average, 1 = maximum, 2 = patches
test_patches = [1, 2, 5, 8, 9, 12, 14]

'''
Enter patches to test for. Can also be entered twice if you
would like twice as much bias on one patch.
'''


def degamma(x):
    x = x / ((2 ** 8) - 1)  # takes 255 and scales it down to one
    x = np.where(x < 0.04045, x / 12.92, ((x + 0.055) / 1.055) ** 2.4)
    x = x * ((2 ** 16) - 1)  # takes one and scales up to 65535, 16 bit color
    return x


def gamma(x):
    # Take 3 long array of color values and gamma them
    return [((colour / 255) ** (1 / 2.4) * 1.055 - 0.055) * 255 for colour in x]


"""
FInds colour correction matrices for list of images
"""


def ccm(Cam, cal_cr_list, cal_cb_list):
    global matrix_selection_types, typenum
    imgs = Cam.imgs
    """
    standard macbeth chart colour values
    """
    m_rgb = np.array([  # these are in RGB
        [116, 81, 67],    # dark skin
        [199, 147, 129],  # light skin
        [91, 122, 156],   # blue sky
        [90, 108, 64],    # foliage
        [130, 128, 176],  # blue flower
        [92, 190, 172],   # bluish green
        [224, 124, 47],   # orange
        [68, 91, 170],    # purplish blue
        [198, 82, 97],    # moderate red
        [94, 58, 106],    # purple
        [159, 189, 63],   # yellow green
        [230, 162, 39],   # orange yellow
        [35, 63, 147],    # blue
        [67, 149, 74],    # green
        [180, 49, 57],    # red
        [238, 198, 20],   # yellow
        [193, 84, 151],   # magenta
        [0, 136, 170],    # cyan (goes out of gamut)
        [245, 245, 243],  # white 9.5
        [200, 202, 202],  # neutral 8
        [161, 163, 163],  # neutral 6.5
        [121, 121, 122],  # neutral 5
        [82, 84, 86],     # neutral 3.5
        [49, 49, 51]      # black 2
    ])
    """
    convert reference colours from srgb to rgb
    """
    m_srgb = degamma(m_rgb)  # now in 16 bit color.

    # Produce array of LAB values for ideal color chart
    m_lab = [colors.RGB_to_LAB(color / 256) for color in m_srgb]

    """
    reorder reference values to match how patches are ordered
    """
    m_srgb = np.array([m_srgb[i::6] for i in range(6)]).reshape((24, 3))
    m_lab = np.array([m_lab[i::6] for i in range(6)]).reshape((24, 3))
    m_rgb = np.array([m_rgb[i::6] for i in range(6)]).reshape((24, 3))
    """
    reformat alsc correction tables or set colour_cals to None if alsc is
    deactivated
    """
    if cal_cr_list is None:
        colour_cals = None
    else:
        colour_cals = {}
        for cr, cb in zip(cal_cr_list, cal_cb_list):
            cr_tab = cr['table']
            cb_tab = cb['table']
            """
            normalise tables so min value is 1
            """
            cr_tab = cr_tab / np.min(cr_tab)
            cb_tab = cb_tab / np.min(cb_tab)
            colour_cals[cr['ct']] = [cr_tab, cb_tab]

    """
    for each image, perform awb and alsc corrections.
    Then calculate the colour correction matrix for that image, recording the
    ccm and the colour tempertaure.
    """
    ccm_tab = {}
    for Img in imgs:
        Cam.log += '\nProcessing image: ' + Img.name
        """
        get macbeth patches with alsc applied if alsc enabled.
        Note: if alsc is disabled then colour_cals will be set to None and no
        the function will simply return the macbeth patches
        """
        r, b, g = get_alsc_patches(Img, colour_cals, grey=False)
        # 256 values for each patch of sRGB values

        """
        do awb
        Note: awb is done by measuring the macbeth chart in the image, rather
        than from the awb calibration. This is done so the awb will be perfect
        and the ccm matrices will be more accurate.
        """
        r_greys, b_greys, g_greys = r[3::4], b[3::4], g[3::4]
        r_g = np.mean(r_greys / g_greys)
        b_g = np.mean(b_greys / g_greys)
        r = r / r_g
        b = b / b_g
        """
        normalise brightness wrt reference macbeth colours and then average
        each channel for each patch
        """
        gain = np.mean(m_srgb) / np.mean((r, g, b))
        Cam.log += '\nGain with respect to standard colours: {:.3f}'.format(gain)
        r = np.mean(gain * r, axis=1)
        b = np.mean(gain * b, axis=1)
        g = np.mean(gain * g, axis=1)
        """
        calculate ccm matrix
        """
        # ==== All of below should in sRGB ===##
        sumde = 0
        ccm = do_ccm(r, g, b, m_srgb)
        # This is the initial guess that our optimisation code works with.
        original_ccm = ccm
        r1 = ccm[0]
        r2 = ccm[1]
        g1 = ccm[3]
        g2 = ccm[4]
        b1 = ccm[6]
        b2 = ccm[7]
        '''
        COLOR MATRIX LOOKS AS BELOW
        R1 R2 R3   Rval     Outr
        G1 G2 G3  *  Gval  =  G
        B1 B2 B3   Bval     B
        Will be optimising 6 elements and working out the third element using 1-r1-r2 = r3
        '''

        x0 = [r1, r2, g1, g2, b1, b2]
        '''
        We use our old CCM as the initial guess for the program to find the
        optimised matrix
        '''
        result = minimize(guess, x0, args=(r, g, b, m_lab), tol=0.01)
        '''
        This produces a color matrix which has the lowest delta E possible,
        based off the input data. Note it is impossible for this to reach
        zero since the input data is imperfect
        '''

        Cam.log += ("\n \n Optimised Matrix Below: \n \n")
        [r1, r2, g1, g2, b1, b2] = result.x
        # The new, optimised color correction matrix values
        optimised_ccm = [r1, r2, (1 - r1 - r2), g1, g2, (1 - g1 - g2), b1, b2, (1 - b1 - b2)]

        # This is the optimised Color Matrix (preserving greys by summing rows up to 1)
        Cam.log += str(optimised_ccm)
        Cam.log += "\n Old Color Correction Matrix Below \n"
        Cam.log += str(ccm)

        formatted_ccm = np.array(original_ccm).reshape((3, 3))

        '''
        below is a whole load of code that then applies the latest color
        matrix, and returns LAB values for color. This can then be used
        to calculate the final delta E
        '''
        optimised_ccm_rgb = []  # Original Color Corrected Matrix RGB / LAB
        optimised_ccm_lab = []

        formatted_optimised_ccm = np.array(optimised_ccm).reshape((3, 3))
        after_gamma_rgb = []
        after_gamma_lab = []

        for RGB in zip(r, g, b):
            ccm_applied_rgb = np.dot(formatted_ccm, (np.array(RGB) / 256))
            optimised_ccm_rgb.append(gamma(ccm_applied_rgb))
            optimised_ccm_lab.append(colors.RGB_to_LAB(ccm_applied_rgb))

            optimised_ccm_applied_rgb = np.dot(formatted_optimised_ccm, np.array(RGB) / 256)
            after_gamma_rgb.append(gamma(optimised_ccm_applied_rgb))
            after_gamma_lab.append(colors.RGB_to_LAB(optimised_ccm_applied_rgb))
        '''
        Gamma After RGB / LAB - not used in calculations, only used for visualisation
        We now want to spit out some data that shows
        how the optimisation has improved the color matrices
        '''
        Cam.log += "Here are the Improvements"

        # CALCULATE WORST CASE delta e
        old_worst_delta_e = 0
        before_average = transform_and_evaluate(formatted_ccm, r, g, b, m_lab)
        new_worst_delta_e = 0
        after_average = transform_and_evaluate(formatted_optimised_ccm, r, g, b, m_lab)
        for i in range(24):
            old_delta_e = deltae(optimised_ccm_lab[i], m_lab[i])  # Current Old Delta E
            new_delta_e = deltae(after_gamma_lab[i], m_lab[i])  # Current New Delta E
            if old_delta_e > old_worst_delta_e:
                old_worst_delta_e = old_delta_e
            if new_delta_e > new_worst_delta_e:
                new_worst_delta_e = new_delta_e

        Cam.log += "Before color correction matrix was optimised, we got an average delta E of " + str(before_average) + " and a maximum delta E of " + str(old_worst_delta_e)
        Cam.log += "After color correction matrix was optimised, we got an average delta E of " + str(after_average) + " and a maximum delta E of " + str(new_worst_delta_e)

        visualise_macbeth_chart(m_rgb, optimised_ccm_rgb, after_gamma_rgb, str(Img.col) + str(matrix_selection_types[typenum]))
        '''
        The program will also save some visualisations of improvements.
        Very pretty to look at. Top rectangle is ideal, Left square is
        before optimisation, right square is after.
        '''

        """
        if a ccm has already been calculated for that temperature then don't
        overwrite but save both. They will then be averaged later on
        """  # Now going to use optimised color matrix, optimised_ccm
        if Img.col in ccm_tab.keys():
            ccm_tab[Img.col].append(optimised_ccm)
        else:
            ccm_tab[Img.col] = [optimised_ccm]
        Cam.log += '\n'

    Cam.log += '\nFinished processing images'
    """
    average any ccms that share a colour temperature
    """
    for k, v in ccm_tab.items():
        tab = np.mean(v, axis=0)
        tab = np.where((10000 * tab) % 1 <= 0.05, tab + 0.00001, tab)
        tab = np.where((10000 * tab) % 1 >= 0.95, tab - 0.00001, tab)
        ccm_tab[k] = list(np.round(tab, 5))
        Cam.log += '\nMatrix calculated for colour temperature of {} K'.format(k)

    """
    return all ccms with respective colour temperature in the correct format,
    sorted by their colour temperature
    """
    sorted_ccms = sorted(ccm_tab.items(), key=lambda kv: kv[0])
    ccms = []
    for i in sorted_ccms:
        ccms.append({
            'ct': i[0],
            'ccm': i[1]
        })
    return ccms


def guess(x0, r, g, b, m_lab):       # provides a method of numerical feedback for the optimisation code
    [r1, r2, g1, g2, b1, b2] = x0
    ccm = np.array([r1, r2, (1 - r1 - r2),
                    g1, g2, (1 - g1 - g2),
                    b1, b2, (1 - b1 - b2)]).reshape((3, 3))  # format the matrix correctly
    return transform_and_evaluate(ccm, r, g, b, m_lab)


def transform_and_evaluate(ccm, r, g, b, m_lab):  # Transforms colors to LAB and applies the correction matrix
    # create list of matrix changed colors
    realrgb = []
    for RGB in zip(r, g, b):
        rgb_post_ccm = np.dot(ccm, np.array(RGB) / 256)  # This is RGB values after the color correction matrix has been applied
        realrgb.append(colors.RGB_to_LAB(rgb_post_ccm))
    # now compare that with m_lab and return numeric result, averaged for each patch
    return (sumde(realrgb, m_lab) / 24)  # returns an average result of delta E


def sumde(listA, listB):
    global typenum, test_patches
    sumde = 0
    maxde = 0
    patchde = []  # Create array of the delta E values for each patch. useful for optimisation of certain patches
    for listA_item, listB_item in zip(listA, listB):
        if maxde < (deltae(listA_item, listB_item)):
            maxde = deltae(listA_item, listB_item)
        patchde.append(deltae(listA_item, listB_item))
        sumde += deltae(listA_item, listB_item)
    '''
    The different options specified at the start allow for
    the maximum to be returned, average or specific patches
    '''
    if typenum == 0:
        return sumde
    if typenum == 1:
        return maxde
    if typenum == 2:
        output = sum([patchde[test_patch] for test_patch in test_patches])
        # Selects only certain patches and returns the output for them
        return output


"""
calculates the ccm for an individual image.
ccms are calculated in rgb space, and are fit by hand. Although it is a 3x3
matrix, each row must add up to 1 in order to conserve greyness, simplifying
calculation.
The initial CCM is calculated in RGB, and then optimised in LAB color space
This simplifies the initial calculation but then gets us the accuracy of
using LAB color space.
"""


def do_ccm(r, g, b, m_srgb):
    rb = r-b
    gb = g-b
    rb_2s = (rb * rb)
    rb_gbs = (rb * gb)
    gb_2s = (gb * gb)

    r_rbs = rb * (m_srgb[..., 0] - b)
    r_gbs = gb * (m_srgb[..., 0] - b)
    g_rbs = rb * (m_srgb[..., 1] - b)
    g_gbs = gb * (m_srgb[..., 1] - b)
    b_rbs = rb * (m_srgb[..., 2] - b)
    b_gbs = gb * (m_srgb[..., 2] - b)

    """
    Obtain least squares fit
    """
    rb_2 = np.sum(rb_2s)
    gb_2 = np.sum(gb_2s)
    rb_gb = np.sum(rb_gbs)
    r_rb = np.sum(r_rbs)
    r_gb = np.sum(r_gbs)
    g_rb = np.sum(g_rbs)
    g_gb = np.sum(g_gbs)
    b_rb = np.sum(b_rbs)
    b_gb = np.sum(b_gbs)

    det = rb_2 * gb_2 - rb_gb * rb_gb

    """
    Raise error if matrix is singular...
    This shouldn't really happen with real data but if it does just take new
    pictures and try again, not much else to be done unfortunately...
    """
    if det < 0.001:
        raise ArithmeticError

    r_a = (gb_2 * r_rb - rb_gb * r_gb) / det
    r_b = (rb_2 * r_gb - rb_gb * r_rb) / det
    """
    Last row can be calculated by knowing the sum must be 1
    """
    r_c = 1 - r_a - r_b

    g_a = (gb_2 * g_rb - rb_gb * g_gb) / det
    g_b = (rb_2 * g_gb - rb_gb * g_rb) / det
    g_c = 1 - g_a - g_b

    b_a = (gb_2 * b_rb - rb_gb * b_gb) / det
    b_b = (rb_2 * b_gb - rb_gb * b_rb) / det
    b_c = 1 - b_a - b_b

    """
    format ccm
    """
    ccm = [r_a, r_b, r_c, g_a, g_b, g_c, b_a, b_b, b_c]

    return ccm


def deltae(colorA, colorB):
    return ((colorA[0] - colorB[0]) ** 2 + (colorA[1] - colorB[1]) ** 2 + (colorA[2] - colorB[2]) ** 2) ** 0.5
    # return ((colorA[1]-colorB[1]) *  * 2 + (colorA[2]-colorB[2]) *  * 2) *  * 0.5
    # UNCOMMENT IF YOU WANT TO NEGLECT LUMINANCE FROM CALCULATION OF DELTA E
